# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import itertools
import os
import traceback
from collections.abc import Collection
from pathlib import Path

import pmb.build
import pmb.chroot
import pmb.config
import pmb.helpers.apk
import pmb.parse.apkindex
from pmb.core import Chroot
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.meta import Cache
from pmb.types import PathString


@Cache("chroot")
def check_min_version(chroot: Chroot = Chroot.native()) -> None:
    """
    Check the minimum apk version, before running it the first time in the
    current session (lifetime of one pmbootstrap call).
    """
    # Skip if apk is not installed yet
    if not (chroot / "sbin/apk").exists():
        logging.debug(
            f"NOTE: Skipped apk version check for chroot '{chroot}'"
            ", because it is not installed yet!"
        )
        return

    installed_pkgs = installed(chroot)

    # Compare
    try:
        version_installed = installed_pkgs["apk-tools"].version
    except KeyError as exception:
        raise NonBugError(
            "ERROR: could not find apk-tools in chroot, run 'pmbootstrap zap' and try again"
        ) from exception

    try:
        pmb.helpers.apk.check_outdated(version_installed)
    except RuntimeError as exception:
        raise NonBugError(
            f"{exception}. Delete your http cache and zap all chroots, then try again: 'pmbootstrap zap -hc'"
        ) from exception


def packages_split_to_add_del(packages: Collection[str]) -> tuple[list[str], list[str]]:
    """
    Sort packages into "to_add" and "to_del" lists depending on their pkgname
    starting with an exclamation mark.

    :param packages: list of pkgnames
    :returns: (to_add, to_del) - tuple of lists of pkgnames, e.g.
              (["hello-world", ...], ["some-conflict-pkg", ...])
    """
    to_add = []
    to_del = []

    for package in packages:
        if package.startswith("!"):
            to_del.append(package.lstrip("!"))
        else:
            to_add.append(package)

    return (to_add, to_del)


def packages_get_locally_built_apks(package_list: Collection[str], arch: Arch) -> list[Path]:
    """
    Iterate over packages and if existing, get paths to locally built packages.
    This is used to force apk to upgrade packages to newer local versions, even
    if the pkgver and pkgrel did not change.

    :param packages: list of pkgnames
    :param arch: architecture that the locally built packages should have
    :returns: Pair of lists, the first is the input packages with local apks removed.
              the second is a list of apk file paths that are valid inside the chroots, e.g.
              ["/mnt/pmbootstrap/packages/x86_64/hello-world-1-r6.apk", ...]
    """
    channels: list[str] = pmb.config.pmaports.all_channels()
    local: list[Path] = []

    packages = set(package_list)

    walked: set[str] = set()
    while len(packages):
        package = packages.pop()
        data_repo = pmb.parse.apkindex.package(package, arch, False)
        if not data_repo:
            continue

        apk_file = f"{data_repo.pkgname}-{data_repo.version}.apk"
        # FIXME: we should know what channel we expect this package to be in
        # this will have weird behaviour if you build gnome-shell for edge and
        # then checkout out the systemd branch... But there isn't
        for channel in channels:
            apk_path = get_context().config.work / "packages" / channel / arch / apk_file
            if apk_path.exists():
                local.append(apk_path)
                break

        # Record all the packages we have visited so far
        walked |= {data_repo.pkgname, package}
        if data_repo.depends:
            # Add all dependencies to the list of packages to check, excluding
            # meta-deps like cmd:* and so:* as well as conflicts (!).
            packages |= (
                set(filter(lambda x: ":" not in x and "!" not in x, data_repo.depends)) - walked
            )

    return local


def install_run_apk(
    to_add: Collection[str], to_add_local: Collection[Path], to_del: Collection[str], chroot: Chroot
) -> None:
    """
    Run apk to add packages, and ensure only the desired packages get
    explicitly marked as installed.

    :param to_add: list of pkgnames to install, without their dependencies
    :param to_add_local: return of packages_get_locally_built_apks()
    :param to_del: list of pkgnames to be deleted, this should be set to
                   conflicting dependencies in any of the packages to be
                   installed or their dependencies (e.g. ["unl0kr"])
    :param chroot: the chroot suffix, e.g. "native" or "rootfs_qemu-amd64"
    """
    context = get_context()
    # Sanitize packages: don't allow '--allow-untrusted' and other options
    # to be passed to apk!
    local_add = [os.fspath(p) for p in to_add_local]
    for package in itertools.chain(to_add, local_add, to_del):
        if package.startswith("-"):
            raise ValueError(f"Invalid package name: {package}")

    commands: list[Collection[PathString]] = [["add", *to_add]]

    # Use a virtual package to mark only the explicitly requested packages as
    # explicitly installed, not the ones in to_add_local
    if to_add_local:
        commands += [
            ["add", "-u", "--virtual", ".pmbootstrap", *local_add],
            ["del", ".pmbootstrap"],
        ]

    if to_del:
        commands += [["del", *to_del]]

    channel = pmb.config.pmaports.read_config()["channel"]
    # There are still some edgecases where we manage to get here while the chroot is not
    # initialized. To not break the build, we initialize it here but print a big warning
    # and a stack trace so hopefully folks report it.
    if not chroot.is_mounted():
        logging.warning(f"({chroot}) chroot not initialized! This is a bug! Please report it.")
        logging.warning(f"({chroot}) initializing the chroot for you...")
        traceback.print_stack(file=logging.logfd)
        pmb.chroot.init(chroot)

    # FIXME: use /mnt/pmb… until MR 2351 is reverted (pmb#2388)
    user_repo: list[PathString] = []
    for channel in pmb.config.pmaports.all_channels():
        user_repo += ["--repository", context.config.work / "packages" / channel]

    for i, command in enumerate(commands):
        command = [*user_repo, *command]

        # Ignore missing repos before initial build (bpo#137)
        if (
            os.getenv("PMB_APK_FORCE_MISSING_REPOSITORIES") == "1"
            or chroot.arch not in Arch.supported_binary()
        ):
            command = ["--force-missing-repositories", *command]

        # Virtual package related commands don't actually install or remove
        # packages, but only mark the right ones as explicitly installed.
        # So only display a progress bar for the "apk add" command which is
        # always the first one we process (i == 0).
        pmb.helpers.apk.run(command, chroot, with_progress=(i == 0))


def install(
    packages: Collection[str], chroot: Chroot, build: bool = True, quiet: bool = False
) -> None:
    """
    Install packages from pmbootstrap's local package index or the pmOS/Alpine
    binary package mirrors. Iterate over all dependencies recursively, and
    build missing packages as necessary.

    :param packages: list of pkgnames to be installed
    :param suffix: the chroot suffix, e.g. "native" or "rootfs_qemu-amd64"
    :param build: automatically build the package, when it does not exist yet
                  or needs to be updated, and it is inside pmaports. For the
                  special case that all packages are expected to be in Alpine's
                  repositories, set this to False for performance optimization.
    """
    arch = chroot.arch
    context = get_context()

    if not packages:
        logging.verbose("pmb.chroot.apk.install called with empty packages list, ignoring")
        return

    # Initialize chroot
    check_min_version(chroot)

    if any(p.startswith("!") for p in packages):
        msg = f"({chroot}) install: packages with '!' are not supported!\n{', '.join(packages)}"
        raise ValueError(msg)

    to_add, to_del = packages_split_to_add_del(packages)

    if build and context.config.build_pkgs_on_install:
        pmb.build.packages(context, to_add, arch)

    to_add_local = packages_get_locally_built_apks(to_add, arch)

    if not quiet:
        logging.info(f"({chroot}) install {' '.join(packages)}")
    install_run_apk(to_add, to_add_local, to_del, chroot)


def installed(suffix: Chroot = Chroot.native()) -> dict[str, pmb.parse.apkindex.ApkindexBlock]:
    """
    Read the list of installed packages (which has almost the same format, as
    an APKINDEX, but with more keys).

    :returns: a dictionary with the following structure:
              { "postmarketos-mkinitfs": ApkindexBlock }

    """
    path = suffix / "lib/apk/db/installed"
    return pmb.parse.apkindex.parse(path, False)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import sys
from collections.abc import Sequence
from pathlib import Path

#
# Exported functions
#
# FIXME (#2324): this sucks, we should re-organise this and not rely on "lifting"
# this functions this way
from pmb.config.file import (
    load as load,
    save as save,
    serialize as serialize,
)
from pmb.config.other import is_systemd_selected as is_systemd_selected
from pmb.config.sudo import which_sudo
from pmb.types import PathString

from . import workdir as workdir
from .init import (
    ask_for_mainline_downstream as ask_for_mainline_downstream,
    require_programs as require_programs,
)

#
# Exported variables (internal configuration)
#
pmb_src: Path = Path(Path(__file__) / "../../..").resolve()
apk_keys_path: Path = pmb_src / "pmb/data/keys"

# apk-tools minimum version
# https://pkgs.alpinelinux.org/packages?name=apk-tools&branch=edge
# Update this frequently to prevent a MITM attack with an outdated version
# (which may contain a vulnerable apk/openssl, and allows an attacker to
# exploit the system!)
apk_tools_min_version = {
    "edge": "3.0.3-r1",
    "v3.23": "3.0.3-r1",
    "v3.22": "2.14.9-r3",
    "v3.21": "2.14.6-r3",
    "v3.20": "2.14.4-r1",
    "v3.19": "2.14.4-r0",
    "v3.18": "2.14.4-r0",
    "v3.17": "2.12.14-r0",
    "v3.16": "2.12.9-r3",
    "v3.15": "2.12.7-r3",
    "v3.14": "2.12.7-r0",
    "v3.13": "2.12.7-r0",
    "v3.12": "2.10.8-r1",
}

# postmarketOS aports compatibility (checked against "version" in pmaports.cfg)
pmaports_min_version = "7"

# Version of the work folder (as asked during 'pmbootstrap init'). Increase
# this number, whenever migration is required and provide the migration code,
# see migrate_work_folder()).
work_version = 8

# Programs that pmbootstrap expects to be available from the host system. Keep
# in sync with README.md, and try to keep the list as small as possible. The
# idea is to run almost everything in Alpine chroots.
required_programs: dict[str, str] = {
    "git": "",
    "kpartx": "",
    "losetup": "",
    "openssl": "",
    "ps": "",
    "tar": "",
    "chroot": "",
    "sh": "",
}


def sudo(cmd: Sequence[PathString]) -> Sequence[PathString]:
    """Adapt a command to run as root."""
    sudo = which_sudo()
    if sudo:
        return [sudo, *cmd]
    else:
        return cmd


defaults: dict[str, PathString] = {
    "cipher": "aes-xts-plain64",
    "config": Path(
        (os.environ.get("XDG_CONFIG_HOME") or os.path.expanduser("~/.config"))
        + "/pmbootstrap_v3.cfg"
    ),
    # A higher value is typically desired, but this can lead to VERY long open
    # times on slower devices due to host systems being MUCH faster than the
    # target device (see issue #429).
    "iter_time": "200",
    "ui": "console",
}

# Whether we're connected to a TTY (which allows things like e.g. printing
# progress bars)
is_interactive = sys.stdout.isatty() and sys.stderr.isatty() and sys.stdin.isatty()


# ANSI escape codes to highlight stdout
styles = {
    "BLUE": "\033[94m",
    "BOLD": "\033[1m",
    "GREEN": "\033[92m",
    "MAGENTA": "\033[95m",
    "RED": "\033[91m",
    "YELLOW": "\033[93m",
    "END": "\033[0m",
}

if "NO_COLOR" in os.environ:
    for style in styles:
        styles[style] = ""

# Supported filesystems and their fstools packages
filesystems = {
    "btrfs": "btrfs-progs",
    "ext2": "e2fsprogs",
    "ext4": "e2fsprogs",
    "f2fs": "f2fs-tools",
    "fat16": "dosfstools",
    "fat32": "dosfstools",
    "xfs": "xfsprogs",
}

# Legacy channels and their new names (pmb#2015)
pmaports_channels_legacy = {"stable": "v20.05", "stable-next": "v21.03"}
#
# CHROOT
#

# Usually the ID for the first user created is 1000. However, we want
# pmbootstrap to work even if the 'user' account inside the chroots has
# another UID, so we force it to be different.
chroot_uid_user = "12345"

# The PATH variable used inside all chroots
chroot_path = ":".join(
    [
        "/usr/lib/ccache/bin",
        "/usr/local/sbin",
        "/usr/local/bin",
        "/usr/sbin:/usr/bin",
        "/sbin",
        "/bin",
    ]
)

# The PATH variable used on the host, to find the "chroot" and "sh"
# executables. As pmbootstrap runs as user, not as root, the location
# for the chroot executable may not be in the PATH (Debian).
host_path = os.environ["PATH"] + ":/usr/sbin/"

# Folders that get mounted inside the chroot
# $WORK gets replaced with get_context().config.work
# $ARCH gets replaced with the chroot architecture (eg. x86_64, armhf)
# $CHANNEL gets replaced with the release channel (e.g. edge, v21.03)
# Use no more than one dir after /mnt/pmbootstrap, see remove_mnt_pmbootstrap.
chroot_mount_bind = {
    "/proc": "/proc",
    "/sys": "/sys",
    "$WORK/cache_apk_$ARCH": "/var/cache/apk",
    "$WORK/cache_appstream/$ARCH/$CHANNEL": "/mnt/appstream-data",
    "$WORK/cache_ccache_$ARCH": "/mnt/pmbootstrap/ccache",
    "$WORK/cache_distfiles": "/var/cache/distfiles",
    "$WORK/cache_git": "/mnt/pmbootstrap/git",
    "$WORK/cache_go": "/mnt/pmbootstrap/go",
    "$WORK/cache_rust": "/mnt/pmbootstrap/rust",
    "$WORK/config_abuild": "/mnt/pmbootstrap/abuild-config",
    "$WORK/config_apk_keys": "/etc/apk/keys",
    "$WORK/cache_sccache": "/mnt/pmbootstrap/sccache",
    "$WORK/images_netboot": "/mnt/pmbootstrap/netboot",
    "$WORK/packages/": "/mnt/pmbootstrap/packages",
}

# Building chroots (all chroots, except for the rootfs_ chroot) get symlinks in
# the "pmos" user's home folder pointing to mountfolders from above.
# Rust packaging is new and still a bit weird in Alpine and postmarketOS. As of
# writing, we only have one package (squeekboard), and use cargo to download
# the source of all dependencies at build time and compile it. Usually, this is
# a no-go, but at least until this is resolved properly, let's cache the
# dependencies and downloads as suggested in "Caching the Cargo home in CI":
# https://doc.rust-lang.org/cargo/guide/cargo-home.html
# Go: cache the directories "go env GOMODCACHE" and "go env GOCACHE" point to,
# to avoid downloading dependencies over and over (GOMODCACHE, similar to the
# rust depends caching described above) and to cache build artifacts (GOCACHE,
# similar to ccache).
chroot_home_symlinks = {
    "/mnt/pmbootstrap/abuild-config": "/home/pmos/.abuild",
    "/mnt/pmbootstrap/ccache": "/home/pmos/.ccache",
    "/mnt/pmbootstrap/go/gocache": "/home/pmos/.cache/go-build",
    "/mnt/pmbootstrap/go/gomodcache": "/home/pmos/go/pkg/mod",
    # "/mnt/pmbootstrap/packages": "/home/pmos/packages/pmos",
    "/mnt/pmbootstrap/rust/git/db": "/home/pmos/.cargo/git/db",
    "/mnt/pmbootstrap/rust/registry/cache": "/home/pmos/.cargo/registry/cache",
    "/mnt/pmbootstrap/rust/registry/index": "/home/pmos/.cargo/registry/index",
    "/mnt/pmbootstrap/sccache": "/home/pmos/.cache/sccache",
}

# Device nodes to be created in each chroot. Syntax for each entry:
# [permissions, type, major, minor, name]
chroot_device_nodes = [
    [666, "c", 1, 3, "null"],
    [666, "c", 1, 5, "zero"],
    [666, "c", 1, 7, "full"],
    [666, "c", 1, 8, "random"],
    [666, "c", 1, 9, "urandom"],
    [666, "c", 5, 0, "tty"],
    [666, "c", 5, 2, "ptmx"],
]

# Age in hours that we keep the APKINDEXes before downloading them again.
# You can force-update them with 'pmbootstrap update'.
apkindex_retention_time = 4


# When chroot is considered outdated (in seconds)
chroot_outdated = 3600 * 24 * 2

# Packages that will be installed in a chroot before it builds packages
# for the first time
# IMPORTANT: the order here matters, it is the order these packages will
# be built in (if needed). abuild must be first!
build_packages = ["abuild", "apk-tools", "build-base", "ccache", "git", "pigz"]

#
# PARSE
#

# Variables belonging to a package or subpackage in APKBUILD files
apkbuild_package_attributes = {
    "pkgdesc": {},
    "depends": {"array": True},
    "provides": {"array": True},
    "provider_priority": {"int": True},
    "install": {"array": True},
    "triggers": {"array": True},
    # Packages can specify soft dependencies in "_pmb_recommends" to be
    # explicitly installed by default, and not implicitly as a hard dependency
    # of the package ("depends"). This makes these apps uninstallable, without
    # removing the meta-package. (#1933). To disable this feature, use:
    # "pmbootstrap install --no-recommends".
    "_pmb_recommends": {"array": True},
    # UI meta-packages can specify groups to which the user must be added
    # to access specific hardware such as LED indicators.
    "_pmb_groups": {"array": True},
    # postmarketos-base, UI and device packages can use _pmb_select to provide
    # additional configuration options in "pmbootstrap init" that allow
    # selecting alternative providers for a virtual APK package.
    "_pmb_select": {"array": True},
    # postmarketos-base and UI meta-packages can define the default package
    # to select during "_pmb_select".
    "_pmb_default": {"array": True},
    # kernel packages can specify one or more defconfig to use when generating
    # the full config
    "_defconfig": {"array": True},
}

# Variables in APKBUILD files that get parsed
apkbuild_attributes = {
    **apkbuild_package_attributes,
    "maintainer": {},
    "arch": {"array": True},
    "depends_dev": {"array": True},
    "makedepends": {"array": True},
    "checkdepends": {"array": True},
    "options": {"array": True},
    "triggers": {"array": True},
    "pkgname": {},
    "pkgrel": {},
    "pkgver": {},
    "sha512sums": {},
    "subpackages": {},
    "url": {},
    # cross-compilers
    "makedepends_build": {"array": True},
    "makedepends_host": {"array": True},
    # kernels
    "_flavor": {},
    "_device": {},
    "_kernver": {},
    "_outdir": {},
    "_config": {},
    # linux-edge
    "_depends_dev": {"array": True},
    # mesa
    "_llvmver": {},
    # Overridden packages
    "_pkgver": {},
    "_pkgrel": {},
    "_pkgname": {},
    # git commit
    "_commit": {},
    "source": {"array": True},
    # gcc
    "_pkgbase": {},
    "_pkgsnap": {},
}

# Valid types for the 'chassis' attribute in deviceinfo
# See https://www.freedesktop.org/software/systemd/man/machine-info.html
deviceinfo_chassis_types = [
    "desktop",
    "laptop",
    "convertible",
    "server",
    "tablet",
    "handset",
    "watch",
    "embedded",
    "vm",
]

#
# INITFS
#
initfs_hook_prefix = "postmarketos-mkinitfs-hook-"
default_ip = "172.16.42.1"


#
# INSTALL
#

# Packages that will be installed inside the native chroot to perform
# the installation to the device.
# util-linux: losetup, fallocate
install_native_packages = ["cryptsetup", "util-linux", "parted"]
install_device_packages = ["postmarketos-base"]

#
# FLASH
#

flash_methods = [
    "0xffff",
    "fastboot",
    "heimdall",
    "mtkclient",
    "none",
    "rkdeveloptool",
    "uuu",
]

# These folders will be mounted at the same location into the native
# chroot, before the flash programs get started.
flash_mount_bind = [
    Path("/sys/bus/usb/devices/"),
    Path("/sys/dev/"),
    Path("/sys/devices/"),
    Path("/dev/bus/usb/"),
]

"""
Flasher abstraction. Allowed variables:

$BOOT: Path to the /boot partition
$DTB: Name of device dtb without .dtb extension
$FLAVOR: Backwards compatibility with old mkinitfs (pma#660)
$IMAGE: Path to the combined boot/rootfs image
$IMAGE_SPLIT_BOOT: Path to the (split) boot image
$IMAGE_SPLIT_ROOT: Path to the (split) rootfs image
$PARTITION_KERNEL: Partition to flash the kernel/boot.img to
$PARTITION_VENDOR_BOOT: Partition to flash the vendor_boot.img to
$PARTITION_ROOTFS: Partition to flash the rootfs to

Fastboot specific: $KERNEL_CMDLINE
Heimdall specific: $PARTITION_INITFS
uuu specific: $UUU_SCRIPT
"""
flashers: dict[str, dict[str, bool | list[str] | dict[str, list[list[str]]]]] = {
    "fastboot": {
        "depends": [],  # pmaports.cfg: supported_fastboot_depends
        "actions": {
            "list_devices": [["fastboot", "devices", "-l"]],
            "flash_rootfs": [["fastboot", "flash", "$PARTITION_ROOTFS", "$IMAGE"]],
            "flash_kernel": [["fastboot", "flash", "$PARTITION_KERNEL", "$BOOT/boot.img$FLAVOR"]],
            "flash_vendorboot": [
                ["fastboot", "flash", "$PARTITION_VENDOR_BOOT", "$BOOT/vendor_boot.img"]
            ],
            "flash_vbmeta": [
                # Generate vbmeta image with "disable verification" flag
                [
                    "avbtool",
                    "make_vbmeta_image",
                    "--flags",
                    "2",
                    "--padding_size",
                    "$FLASH_PAGESIZE",
                    "--output",
                    "/vbmeta.img",
                ],
                ["fastboot", "flash", "$PARTITION_VBMETA", "/vbmeta.img"],
                ["rm", "-f", "/vbmeta.img"],
            ],
            "flash_dtbo": [["fastboot", "flash", "$PARTITION_DTBO", "$BOOT/dtbo.img"]],
            "boot": [["fastboot", "--cmdline", "$KERNEL_CMDLINE", "boot", "$BOOT/boot.img$FLAVOR"]],
            # Like boot, but stage vendor_boot first
            "boot_gki": [
                ["fastboot", "stage", "$BOOT/vendor_boot.img"],
                ["fastboot", "boot", "$BOOT/boot.img$FLAVOR"],
            ],
            "flash_lk2nd": [["fastboot", "flash", "$PARTITION_KERNEL", "$BOOT/lk2nd.img"]],
        },
    },
    # Some devices provide Fastboot but using Android boot images is not
    # practical for them (e.g. because they support booting from FAT32
    # partitions directly and/or the Android boot partition is too small).
    # This can be implemented using --split (separate image files for boot and
    # rootfs).
    # This flasher allows flashing the split image files using Fastboot.
    "fastboot-bootpart": {
        "split": True,
        "depends": ["android-tools"],
        "actions": {
            "list_devices": [["fastboot", "devices", "-l"]],
            "flash_rootfs": [["fastboot", "flash", "$PARTITION_ROOTFS", "$IMAGE_SPLIT_ROOT"]],
            "flash_kernel": [["fastboot", "flash", "$PARTITION_KERNEL", "$IMAGE_SPLIT_BOOT"]],
            "flash_boot": [["fastboot", "flash", "boot", "$BOOT/boot.img$FLAVOR"]],
            "boot": [["fastboot", "--cmdline", "$KERNEL_CMDLINE", "boot", "$BOOT/boot.img$FLAVOR"]],
        },
    },
    # Some Samsung devices need the initramfs to be baked into the kernel (e.g.
    # i9070, i9100). We want the initramfs to be generated after the kernel was
    # built, so we put the real initramfs on another partition (e.g. RECOVERY)
    # and load it from the initramfs in the kernel. This method is called
    # "isorec" (isolated recovery), a term coined by Lanchon.
    "heimdall-isorec": {
        "depends": ["heimdall"],
        "actions": {
            "list_devices": [["heimdall", "detect"]],
            "flash_rootfs": [
                ["heimdall", "flash", "--wait", "--$PARTITION_ROOTFS", "$IMAGE"],
            ],
            "flash_kernel": [
                [
                    "heimdall_flash_isorec_kernel.sh",
                    "$BOOT/initramfs$FLAVOR",
                    "$PARTITION_INITFS",
                    "$BOOT/vmlinuz$FLAVOR",
                    "$PARTITION_KERNEL",
                    "$BOOT/$DTB",
                ]
            ],
        },
    },
    # Some Samsung devices need a 'boot.img' file, just like the one generated
    # fastboot compatible devices. Example: s7562, n7100
    "heimdall-bootimg": {
        "depends": [],  # pmaports.cfg: supported_heimdall_depends
        "actions": {
            "list_devices": [["heimdall", "detect"]],
            "flash_rootfs": [
                [
                    "heimdall",
                    "flash",
                    "--wait",
                    "--$PARTITION_ROOTFS",
                    "$IMAGE",
                    "$NO_REBOOT",
                    "$RESUME",
                ],
            ],
            "flash_kernel": [
                [
                    "heimdall",
                    "flash",
                    "--wait",
                    "--$PARTITION_KERNEL",
                    "$BOOT/boot.img$FLAVOR",
                    "$NO_REBOOT",
                    "$RESUME",
                ],
            ],
            "flash_vbmeta": [
                [
                    "avbtool",
                    "make_vbmeta_image",
                    "--flags",
                    "2",
                    "--padding_size",
                    "$FLASH_PAGESIZE",
                    "--output",
                    "/vbmeta.img",
                ],
                [
                    "heimdall",
                    "flash",
                    "--$PARTITION_VBMETA",
                    "/vbmeta.img",
                    "$NO_REBOOT",
                    "$RESUME",
                ],
                ["rm", "-f", "/vbmeta.img"],
            ],
            "flash_lk2nd": [
                [
                    "heimdall",
                    "flash",
                    "--wait",
                    "--$PARTITION_KERNEL",
                    "$BOOT/lk2nd.img",
                    "$NO_REBOOT",
                    "$RESUME",
                ],
            ],
        },
    },
    "adb": {
        "depends": ["android-tools"],
        "actions": {
            "list_devices": [["adb", "-P", "5038", "devices"]],
            "sideload": [
                ["echo", "< wait for any device >"],
                ["adb", "-P", "5038", "wait-for-usb-sideload"],
                ["adb", "-P", "5038", "sideload", "$RECOVERY_ZIP"],
            ],
        },
    },
    "uuu": {
        "depends": ["nxp-mfgtools-uuu"],
        "actions": {
            "flash_rootfs": [
                # There's a bug(?) in uuu where it clobbers the path in the cmd
                # script if the script is not in pwd...
                ["cp", "$UUU_SCRIPT", "./flash_script.lst"],
                ["uuu", "flash_script.lst"],
            ],
        },
    },
    "rkdeveloptool": {
        "split": True,
        "depends": ["rkdeveloptool"],
        "actions": {
            "list_devices": [["rkdeveloptool", "list"]],
            "flash_rootfs": [
                ["rkdeveloptool", "write-partition", "$PARTITION_ROOTFS", "$IMAGE_SPLIT_ROOT"]
            ],
            "flash_kernel": [
                ["rkdeveloptool", "write-partition", "$PARTITION_KERNEL", "$IMAGE_SPLIT_BOOT"]
            ],
        },
    },
    "mtkclient": {
        "depends": ["mtkclient"],
        "actions": {
            "flash_rootfs": [["mtk", "w", "$PARTITION_ROOTFS", "$IMAGE"]],
            "flash_kernel": [["mtk", "w", "$PARTITION_KERNEL", "$BOOT/boot.img$FLAVOR"]],
            "flash_vbmeta": [
                # Generate vbmeta image with "disable verification" flag
                [
                    "avbtool",
                    "make_vbmeta_image",
                    "--flags",
                    "2",
                    "--padding_size",
                    "$FLASH_PAGESIZE",
                    "--output",
                    "/vbmeta.img",
                ],
                ["mtk", "w", "$PARTITION_VBMETA", "/vbmeta.img"],
                ["rm", "-f", "/vbmeta.img"],
            ],
            "flash_dtbo": [["mtk", "w", "$PARTITION_DTBO", "$BOOT/dtbo.img"]],
            "flash_lk2nd": [["mtk", "w", "$PARTITION_KERNEL", "$BOOT/lk2nd.img"]],
        },
    },
}

#
# GIT
#
git_repos = {
    "aports_upstream": [
        "https://gitlab.alpinelinux.org/alpine/aports.git",
        "git@gitlab.alpinelinux.org:alpine/aports.git",
    ],
    "pmaports": [
        "https://gitlab.postmarketos.org/postmarketOS/pmaports.git",
        "git@gitlab.postmarketos.org:postmarketOS/pmaports.git",
    ],
}

#
# APORTGEN
#

# Use a deterministic mirror URL instead of CDN for aportgen. Otherwise we may
# generate a pmaport that wraps an apk from Alpine (e.g. musl-armv7) locally
# with one up-to-date mirror given by the CDN. But then the build will fail if
# CDN picks an outdated mirror for CI or BPO.
aportgen_mirror_alpine = "http://dl-4.alpinelinux.org/alpine/"

#
# NEWAPKBUILD
# Options passed through to the "newapkbuild" command from Alpine Linux. They
# are duplicated here, so we can use Python's argparse for argument parsing and
# help page display. The -f (force) flag is not defined here, as we use that in
# the Python code only and don't pass it through.
#
newapkbuild_arguments_strings = [
    ["-n", "pkgname", "set package name (only use with SRCURL)"],
    ["-d", "pkgdesc", "set package description"],
    ["-l", "license", "set package license identifier from <https://spdx.org/licenses/>"],
    ["-u", "url", "set package URL"],
]
newapkbuild_arguments_switches_pkgtypes = [
    ["-a", "autotools", "create autotools package (use ./configure ...)"],
    ["-C", "cmake", "create CMake package (assume cmake/ is there)"],
    ["-m", "meson", "create meson package (assume meson.build is there)"],
    ["-p", "perl", "create perl package (assume Makefile.PL is there)"],
    ["-y", "python", "create python package (assume setup.py is there)"],
    ["-e", "python_gpep517", "create python package (assume pyproject.toml is there)"],
    ["-r", "rust", "create rust package (assume Cargo.toml is there)"],
]
newapkbuild_arguments_switches_other = [
    ["-s", "sourceforge", "use sourceforge source URL"],
    ["-c", "copy_samples", "copy a sample init.d, conf.d and install script"],
]

#
# CI
#
# Valid options for 'pmbootstrap ci', see https://postmarketos.org/pmb-ci
ci_valid_options = ["native", "slow", "with-dot-git"]



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import glob
import os
from pathlib import Path

import pmb.build.other
import pmb.chroot
import pmb.config.pmaports
import pmb.config.workdir
import pmb.helpers.apk
import pmb.helpers.cli
import pmb.helpers.mount
import pmb.helpers.pmaports
import pmb.helpers.run
import pmb.parse.apkindex
from pmb.core import Chroot
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import CommandFailedError, NonBugError


def del_chroot(path: Path, confirm: bool = True, dry: bool = False) -> None:
    if confirm and not pmb.helpers.cli.confirm(f"Remove {path}?"):
        return
    if dry:
        return

    # Safety first!
    assert path.is_absolute()
    assert path.is_relative_to(get_context().config.work)

    # umount_all() will throw if any mount under path fails to unmount
    pmb.helpers.mount.umount_all(path)

    pmb.helpers.run.root(["rm", "-rf", path])


def zap(
    confirm: bool = True,
    dry: bool = False,
    pkgs_local: bool = False,
    http: bool = False,
    pkgs_local_mismatch: bool = False,
    pkgs_online_mismatch: bool = False,
    distfiles: bool = False,
    rust: bool = False,
    netboot: bool = False,
) -> None:
    """
    Shutdown everything inside the chroots (e.g. adb), umount
    everything and then safely remove folders from the work-directory.

    :param dry: Only show what would be deleted, do not delete for real
    :param pkgs_local: Remove *all* self-compiled packages (!)
    :param http: Clear the http cache (used e.g. for the initial apk download)
    :param pkgs_local_mismatch: Remove the packages that have
        a different version compared to what is in the aports folder.
    :param pkgs_online_mismatch: Clean out outdated binary packages
        downloaded from mirrors (e.g. from Alpine)
    :param distfiles: Clear the downloaded files cache
    :param rust: Remove rust related caches
    :param netboot: Remove images for netboot

    NOTE: This function gets called in pmb/config/init.py, with only get_context().config.work
    and args.device set!
    """
    # Delete packages with a different version compared to aports,
    # then re-index
    if pkgs_local_mismatch:
        zap_pkgs_local_mismatch(confirm, dry)

    # Delete outdated binary packages
    if pkgs_online_mismatch:
        zap_pkgs_online_mismatch(confirm, dry)

    try:
        pmb.chroot.shutdown()
    except CommandFailedError as exception:
        raise NonBugError(
            f"Failed to shut down all chroots. Ensure they're not doing any work and you're not chrooted into any. ({exception})"
        ) from exception

    # Deletion patterns for folders inside get_context().config.work
    patterns = []
    if pkgs_local:
        patterns += ["packages"]
    if http:
        patterns += ["cache_http"]
    if distfiles:
        patterns += ["cache_distfiles"]
    if rust:
        patterns += ["cache_rust"]
    if netboot:
        patterns += ["images_netboot"]

    for chroot in Chroot.glob():
        del_chroot(chroot, confirm, dry)

    # Delete everything matching the patterns
    for pattern in patterns:
        logging.debug(f"Deleting {pattern}")
        pattern = os.path.realpath(f"{get_context().config.work}/{pattern}")
        matches = glob.glob(pattern)
        for match in matches:
            if not confirm or pmb.helpers.cli.confirm(f"Remove {match}?"):
                logging.info(f"% rm -rf {match}")
                if not dry:
                    pmb.helpers.run.root(["rm", "-rf", match])

    # Remove config init dates for deleted chroots
    pmb.config.workdir.clean()

    # Chroots were zapped, so no repo lists exist anymore
    pmb.helpers.apk.update_repository_list.cache_clear()
    # Let chroot.init be called again
    pmb.chroot.init.cache_clear()

    # Print amount of cleaned up space
    if dry:
        logging.info("Dry run: nothing has been deleted")


def zap_pkgs_local_mismatch(confirm: bool = True, dry: bool = False) -> None:
    channel = pmb.config.pmaports.read_config()["channel"]
    if not os.path.exists(f"{get_context().config.work}/packages/{channel}"):
        return

    question = (
        "Remove binary packages that are newer than the corresponding"
        f" pmaports (channel '{channel}')?"
    )
    if confirm and not pmb.helpers.cli.confirm(question):
        return

    reindex = False
    for apkindex_path in (get_context().config.work / "packages" / channel).glob(
        "*/APKINDEX.tar.gz"
    ):
        # Delete packages without same version in aports
        blocks = pmb.parse.apkindex.parse_blocks(apkindex_path)
        for block in blocks:
            pkgname = block.pkgname
            origin = block.origin
            version = block.version
            arch = block.arch

            # Apk path
            apk_path_short = f"{arch}/{pkgname}-{version}.apk"
            apk_path = f"{get_context().config.work}/packages/{channel}/{apk_path_short}"
            if not os.path.exists(apk_path):
                logging.info(f"WARNING: Package mentioned in index not found: {apk_path_short}")
                continue

            if origin is None:
                raise RuntimeError("Can't handle virtual packages")

            # Aport path
            aport_path = pmb.helpers.pmaports.find_optional(origin)
            if not aport_path:
                logging.info(f"% rm {apk_path_short} ({origin} aport not found)")
                if not dry:
                    pmb.helpers.run.root(["rm", apk_path])
                    reindex = True
                continue

            # Clear out any binary apks that do not match what is in aports
            apkbuild = pmb.parse.apkbuild(aport_path)
            version_aport = f"{apkbuild['pkgver']}-r{apkbuild['pkgrel']}"
            if version != version_aport:
                logging.info(f"% rm {apk_path_short} ({origin} aport: {version_aport})")
                if not dry:
                    pmb.helpers.run.root(["rm", apk_path])
                    reindex = True

    if reindex:
        pmb.build.other.index_repo()


def zap_pkgs_online_mismatch(confirm: bool = True, dry: bool = False) -> None:
    # Check whether we need to do anything
    paths = list(get_context().config.work.glob("cache_apk_*"))
    if not len(paths):
        return
    if confirm and not pmb.helpers.cli.confirm("Remove outdated binary packages?"):
        return

    # Iterate over existing apk caches
    clean_native = False
    for path in paths:
        arch = Arch.from_str(path.name.split("_", 2)[2])

        if not dry:
            if arch == Arch.native():
                clean_native = True
                continue
            pmb.helpers.apk.cache_clean(arch)

    # pmb#2672: With apkv3, APKINDEX files get deleted when running "apk cache
    # clean". This is unexpected and needs to be investigated. Meanwhile run
    # cache_clean() on the native arch last, as it needs the APKINDEX for the
    # Alpine main repo of the native arch (even if cleaning other arches) in
    # order to set up the chroot where we run "apk cache clean". This prevents
    # crashing in "pmbootstrap zap -a" and "pmbootstrap zap -o".
    if clean_native and not dry:
        pmb.helpers.apk.cache_clean(Arch.native())



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import fnmatch
import re
from pathlib import Path

import pmb.helpers.git
import pmb.helpers.run
from pmb.core import Chroot
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError


def indent_size(line: str) -> int:
    """Number of spaces at the beginning of a string."""
    matches = re.findall(r"^[ ]*", line)
    if len(matches) == 1:
        return len(matches[0])
    return 0


def format_function(name: str, body: str, remove_indent: int = 4) -> str:
    """
    Format the body of a shell function passed to rewrite() below, so it fits
    the format of the original APKBUILD.

    :param remove_indent: Maximum number of spaces to remove from the
        beginning of each line of the function body.
    """
    tab_width = 4
    ret = ""
    lines = body.split("\n")
    for i in range(len(lines)):
        line = lines[i]
        if not line.strip() and (not ret or i == len(lines) - 1):
            continue

        # Remove indent
        spaces = min(indent_size(line), remove_indent)
        line = line[spaces:]

        # Convert spaces to tabs
        spaces = indent_size(line)
        tabs = int(spaces / tab_width)
        line = ("\t" * tabs) + line[spaces:]

        ret += line + "\n"
    return name + "() {\n" + ret + "}\n"


def rewrite(
    pkgname: str,
    path_original: Path | str | None = None,
    fields: dict[str, str] = {},
    replace_pkgname: str | None = None,
    replace_functions: dict[str, str | None] = {},
    replace_simple: dict[str, str | None] = {},
    below_header: str = "",
    remove_indent: int = 4,
) -> None:
    """
    Append a header to $WORK/aportgen/APKBUILD, delete maintainer/contributor
    lines (so they won't be bugged with issues regarding our generated aports),
    and add reference to the original aport.

    :param path_original: The original path of the automatically generated
        aport.
    :param fields: key-value pairs of fields that shall be changed in the
        APKBUILD. For example: {"pkgdesc": "my new package", "subpkgs": ""}
    :param replace_pkgname: When set, $pkgname gets replaced with that string
        in every line.
    :param replace_functions: Function names and new bodies, for example:
        {"build": "return 0"}
        The body can also be None (deletes the function)
    :param replace_simple: Lines that fnmatch the pattern, get
        replaced/deleted. Example: {"*test*": "# test", "*mv test.bin*": None}
    :param below_header: String that gets directly placed below the header.
    :param remove_indent: Number of spaces to remove from function body
        provided to replace_functions.

    """
    # Header
    if path_original:
        lines_new = [
            "# Automatically generated aport, do not edit!\n",
            f"# Generator: pmbootstrap aportgen {pkgname}\n",
            f"# Based on: {path_original}\n",
        ]
    else:
        lines_new = [
            "# Forked from Alpine INSERT-REASON-HERE (CHANGEME!)\n",
        ]

    line: str | None
    if below_header:
        for line in below_header.split("\n"):
            if not line[:8].strip():
                line = line[8:]
            lines_new += line.rstrip() + "\n"

    # Copy/modify lines, skip Contributor
    path = get_context().config.work / "aportgen/APKBUILD"
    with open(path, "r+", encoding="utf-8") as handle:
        skip_in_func = False
        have_maintainer = False
        for line in handle:
            # Clear maintainer line
            if line.startswith(("# Maintainer", "maintainer=")):
                if have_maintainer:
                    line = ""  # APKBUILDs may use the variable + comment
                else:
                    line = 'maintainer="YOUR NAME <EMAIL@ADDRESS> (CHANGEME!)"\n'
                    have_maintainer = True

            # Contributor
            if line.startswith("# Contributor"):
                continue

            # Replace functions
            if skip_in_func:
                if line.startswith("}"):
                    skip_in_func = False
                continue
            else:
                for func, body in replace_functions.items():
                    if line.startswith(func + "() {"):
                        skip_in_func = True
                        if body:
                            lines_new += format_function(func, body, remove_indent=remove_indent)
                        break
                if skip_in_func:
                    continue

            # Replace fields
            for key, value in fields.items():
                if line.startswith(key + "="):
                    if value:
                        if key in ["pkgname", "pkgver", "pkgrel"]:
                            # No quotes to avoid lint error
                            line = f"{key}={value}\n"
                        else:
                            line = f'{key}="{value}"\n'
                    else:
                        # Remove line without value to avoid lint error
                        line = ""
                    break

            # Replace $pkgname
            if replace_pkgname and "$pkgname" in line:
                line = line.replace("$pkgname", replace_pkgname)

            # Replace simple
            for pattern, replacement in replace_simple.items():
                if fnmatch.fnmatch(line, pattern + "\n"):
                    if line := replacement:
                        line += "\n"
                    break
            if line is None:
                continue

            lines_new.append(line)

        # Write back
        handle.seek(0)
        handle.write("".join(lines_new))
        handle.truncate()


def get_upstream_aport(pkgname: str, arch: Arch | None = None, retain_branch: bool = False) -> Path:
    """
    Perform a git checkout of Alpine's aports and get the path to the aport.

    :param pkgname: package name
    :param arch: Alpine architecture (e.g. "armhf"), defaults to native arch
    :returns: absolute path on disk where the Alpine aport is checked out
              example: /opt/pmbootstrap_work/cache_git/aports/upstream/main/gcc
    """
    # APKBUILD
    pmb.helpers.git.clone("aports_upstream")
    aports_upstream_path = get_context().config.work / "cache_git/aports_upstream"

    if retain_branch:
        logging.info("Not changing aports branch as --fork-alpine-retain-branch was used.")
    else:
        # Checkout branch
        channel_cfg = pmb.config.pmaports.read_config_channel()
        branch = channel_cfg["branch_aports"]
        logging.info(f"Checkout aports.git branch: {branch}")
        if pmb.helpers.run.user(["git", "switch", branch], aports_upstream_path, check=False):
            logging.info("NOTE: run 'pmbootstrap pull' and try again")
            logging.info(
                "NOTE: if it still fails, your aports.git was cloned with"
                " an older version of pmbootstrap, as shallow clone."
                " Unshallow it, or remove it and let pmbootstrap clone it"
                f" again: {aports_upstream_path}"
            )
            raise RuntimeError("Branch switch failed.")

    # Search package
    paths = list(aports_upstream_path.glob(f"*/{pkgname}"))
    if len(paths) > 1:
        raise RuntimeError("Package " + pkgname + " found in multiple aports subfolders.")
    elif len(paths) == 0:
        raise NonBugError(
            f"Package {pkgname} not found in alpine aports repository. Maybe it's a subpackage of a different origin package?"
        )
    aport_path = paths[0]

    # Parse APKBUILD
    apkbuild = pmb.parse.apkbuild(aport_path, check_pkgname=False)
    apkbuild_version = apkbuild["pkgver"] + "-r" + apkbuild["pkgrel"]

    # Binary package
    split = aport_path.parts
    repo = split[-2]
    pkgname = split[-1]
    # Update or create APKINDEX for relevant arch so we know it exists and is recent.
    pmb.helpers.repo.update(arch)
    index_path = pmb.helpers.repo.alpine_apkindex_path(repo, arch)
    package = pmb.parse.apkindex.package(pkgname, indexes=[index_path], arch=arch)

    if package is None:
        raise RuntimeError(f"Couldn't find {pkgname} in APKINDEX!")

    # Compare version (return when equal)
    compare = pmb.parse.version.compare(apkbuild_version, package.version)

    # APKBUILD > binary: this is fine
    if compare == 1:
        logging.info(
            f"NOTE: {pkgname} {arch} binary package has a lower"
            f" version {package.version} than the APKBUILD"
            f" {apkbuild_version}"
        )
        return aport_path

    # APKBUILD < binary: aports.git is outdated
    if compare == -1:
        logging.warning(
            "WARNING: Package '" + pkgname + "' has a lower version in"
            " local checkout of Alpine's aports ("
            + apkbuild_version
            + ") compared to Alpine's binary package ("
            + package.version
            + ")!"
        )
        logging.info("NOTE: You can update your local checkout with: 'pmbootstrap pull'")

    return aport_path


def prepare_tempdir() -> Path:
    """
    Prepare a temporary directory to do aportgen-related operations within.

    :returns: Path to a temporary directory for aportgen to work within.
    """
    # Prepare aportgen tempdir inside and outside of chroot
    chroot = Chroot.native()
    pmb.chroot.init(chroot)
    tempdir = Path("/tmp/aportgen")
    aportgen = get_context().config.work / "aportgen"
    pmb.chroot.root(["rm", "-rf", tempdir], chroot)
    pmb.helpers.run.user(["mkdir", "-p", aportgen, chroot / tempdir])

    return tempdir


def generate_checksums(tempdir: Path, apkbuild_path: Path) -> None:
    """
    Generate new checksums for a given APKBUILD.

    :param tempdir: Temporary directory as provided by prepare_tempdir().
    :param apkbuild_path: APKBUILD to generate new checksums for.
    """
    aportgen = get_context().config.work / "aportgen"

    pmb.build.init_abuild_minimal()
    pmb.chroot.root(["chown", "-R", "pmos:pmos", tempdir])
    pmb.chroot.user(["abuild", "checksum"], working_dir=tempdir)
    pmb.helpers.run.user(["cp", apkbuild_path, aportgen])



# Copyright 2026 Oliver Smith, Paul Adam
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.chroot
import pmb.chroot.apk
from pmb.core import Chroot
from pmb.helpers import logging
from pmb.types import Arch, RunOutputTypeDefault


def stats(arch: Arch | None) -> None:
    # Chroot suffix
    chroot = Chroot.buildroot(arch or Arch.native())

    pmb.chroot.init(chroot)

    # Install ccache and display stats
    pmb.chroot.apk.install(["ccache"], chroot)
    logging.info(f"({chroot}) % ccache -s")
    pmb.chroot.user(["ccache", "-s"], chroot, output=RunOutputTypeDefault.STDOUT)



# Copyright 2023 Robert Yang
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import re
from pathlib import Path

import pmb.aportgen
import pmb.aportgen.core
import pmb.build
import pmb.chroot
import pmb.helpers
import pmb.helpers.mount
import pmb.helpers.pmaports
import pmb.parse
from pmb.core import Chroot
from pmb.core.arch import Arch
from pmb.core.context import Context, get_context
from pmb.core.pkgrepo import pkgrepo_relative_path
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.types import Env, PathString


def match_kbuild_out(word: str) -> str | None:
    """
    Look for paths in the following formats:
      "<prefix>/<kbuild_out>/arch/<arch>/boot"
      "<prefix>/<kbuild_out>/include/config/kernel.release"

    :param word: space separated string cut out from a line from an APKBUILD
                 function body that might be the kbuild output path
    :returns: kernel build output directory.
              empty string when a separate build output directory isn't used.
              None, when no output directory is found.
    """
    prefix = '^\\"?\\$({?builddir}?|{?srcdir}?)\\"?/'
    kbuild_out = "(.*\\/)*"

    postfix = '(arch\\/.*\\/boot.*)\\"?$'
    match = re.match(prefix + kbuild_out + postfix, word)

    if match is None:
        postfix = '(include\\/config\\/kernel\\.release)\\"?$'
        match = re.match(prefix + kbuild_out + postfix, word)

    if match is None:
        return None

    groups = match.groups()
    if len(groups) != 3:
        return None

    logging.debug("word = " + str(word))
    logging.debug("regex match groups = " + str(groups))
    out_dir = groups[1]
    return "" if out_dir is None else out_dir.strip("/")


def find_kbuild_output_dir(function_body: list[str]) -> str:
    """
    Guess what the kernel build output directory is.

    Parses each line of the function word by word, looking for paths which
    contain the kbuild output directory.

    :param function_body: contents of a function from the kernel APKBUILD
    :returns: kbuild output dir
              None, when output dir is not found
    """
    guesses = []
    for line in function_body:
        for item in line.split():
            # Guess that any APKBUILD using downstreamkernel_package
            # uses the default kbuild out directory.
            if item == "downstreamkernel_package":
                guesses.append("")
                break
            kbuild_out = match_kbuild_out(item)
            if kbuild_out is not None:
                guesses.append(kbuild_out)
                break

    # Check if guesses are all the same
    it = iter(guesses)
    first = next(it, None)
    if first is None:
        raise RuntimeError(
            "Couldn't find a kbuild out directory. Is your APKBUILD messed up? If not, then "
            "consider adjusting the patterns in pmb/build/envkernel.py to work with your APKBUILD, "
            "or submit an issue."
        )
    if all(first == rest for rest in it):
        return first
    raise RuntimeError(
        "Multiple kbuild out directories found. Can you modify your APKBUILD so it only has one "
        "output path? If you can't resolve it, please open an issue."
    )


def modify_apkbuild(pkgname: str, aport: Path) -> None:
    """Modify kernel APKBUILD to package build output from envkernel.sh."""
    work = get_context().config.work
    apkbuild_path = aport / "APKBUILD"
    apkbuild = pmb.parse.apkbuild(apkbuild_path)
    if os.path.exists(work / "aportgen"):
        pmb.helpers.run.user(["rm", "-r", work / "aportgen"])

    pmb.helpers.run.user(["mkdir", work / "aportgen"])
    pmb.helpers.run.user(["cp", "-r", apkbuild_path, work / "aportgen"])

    pkgver = pmb.build._package.get_pkgver(apkbuild["pkgver"], original_source=False)
    fields = {
        "pkgver": pkgver,
        "pkgrel": "0",
        "subpackages": "",
        "builddir": "/home/pmos/build/src",
    }

    pmb.aportgen.core.rewrite(pkgname, apkbuild_path, fields=fields)


def overmount_makefile(context: Context, kbuild_out: str) -> None:
    makefile = context.config.work / "tmp_kernel.mk"
    makefile.open("w").write("""
# Automatically generated by pmbootstrap to use relative include paths
include ../Makefile
""")

    makefile_target = Chroot.native() / "mnt/linux" / kbuild_out / "Makefile"

    logging.info("Overmounting Makefile to use relative include paths")
    if not pmb.helpers.mount.ismount(makefile_target):
        pmb.helpers.mount.bind_file(makefile, makefile_target)


def run_abuild(
    context: Context,
    pkgname: str,
    arch: Arch,
    pmaports_path: Path,
    apkbuild_path: Path,
    kbuild_out: str,
) -> None:
    """
    Prepare build environment and run abuild.

    :param pkgname: package name of a linux kernel aport
    :param arch: architecture for the kernel
    :param pmaports_path: path to the aports dir this package is part of
    :param apkbuild_path: path to APKBUILD of the kernel aport
    :param kbuild_out: kernel build system output sub-directory
    """
    chroot = Chroot.native()
    build_path = Path("/home/pmos/build")
    kbuild_out_source = f"/mnt/linux/{kbuild_out}"

    # If the kernel was cross-compiled on the host rather than with the envkernel
    # helper, we can still use the envkernel logic to package the artifacts for
    # development, making it easy to quickly sideload a new kernel or pmbootstrap
    # to create a boot image.

    pmb.helpers.mount.bind(Path("."), chroot / "mnt/linux")

    if kbuild_out:
        overmount_makefile(context, kbuild_out)

    if not os.path.exists(chroot / kbuild_out_source):
        raise RuntimeError(
            "No '.output' dir found in your kernel source dir. Compile the "
            f"{context.config.device} kernel first and then try again. See "
            "https://postmarketos.org/envkernel for details. If building on your host and only "
            "using --envkernel for packaging, make sure you have O=.output as an argument to make."
        )

    # Create working directory for abuild
    pmb.build.copy_to_buildpath(pkgname)

    # FIXME: duplicated from pmb.build._package.run_aports()
    # This is needed to set up the package output directory for
    # abuild and shouldn't really be done here.
    channel = pmb.config.pmaports.read_config(pmaports_path)["channel"]
    print(f"Building for channel: {channel}")
    pkgdir = context.config.work / "packages" / channel
    if not pkgdir.exists():
        pmb.helpers.run.root(["mkdir", "-p", pkgdir])
        pmb.helpers.run.root(
            [
                "chown",
                "-R",
                f"{pmb.config.chroot_uid_user}:{pmb.config.chroot_uid_user}",
                pkgdir.parent,
            ]
        )

    pmb.chroot.rootm(
        [
            ["mkdir", "-p", "/home/pmos/packages"],
            ["rm", "-f", "/home/pmos/packages/pmos"],
            ["ln", "-sf", f"/mnt/pmbootstrap/packages/{channel}", "/home/pmos/packages/pmos"],
        ],
        chroot,
    )

    # Create symlink from abuild working directory to envkernel build directory
    pmb.chroot.root(["ln", "-sf", kbuild_out_source, build_path / "src"])

    cmd: list[PathString] = ["cp", apkbuild_path, chroot / build_path / "APKBUILD"]
    pmb.helpers.run.root(cmd)

    # Create the apk package
    env: Env = {
        "CARCH": str(arch),
        "CHOST": str(arch),
        "CBUILD": str(Arch.native()),
        "SUDO_APK": "abuild-apk --no-progress",
    }
    cmd = ["abuild", "rootpkg"]
    pmb.chroot.user(cmd, working_dir=build_path, env=env)

    # Clean up bindmount
    pmb.helpers.mount.umount_all(chroot / "mnt/linux")

    # Clean up symlinks
    if (
        kbuild_out != ""
        and os.path.islink(chroot / "mnt/linux" / kbuild_out)
        and os.path.lexists(chroot / "mnt/linux" / kbuild_out)
    ):
        pmb.chroot.root(["rm", Path("/mnt/linux", kbuild_out)])
    pmb.chroot.root(["rm", build_path / "src"])


def package_kernel(packages: list[str]) -> None:
    """Frontend for 'pmbootstrap build --envkernel': creates a package from envkernel output."""
    pkgname = packages[0]
    if len(packages) > 1 or not pkgname.startswith("linux-"):
        raise NonBugError("--envkernel needs exactly one linux-* package as argument.")

    aport = pmb.helpers.pmaports.find(pkgname)
    context = get_context()

    modify_apkbuild(pkgname, aport)
    apkbuild_path = context.config.work / "aportgen/APKBUILD"

    arch = pmb.parse.deviceinfo().arch
    apkbuild = pmb.parse.apkbuild(apkbuild_path, check_pkgname=False)
    if apkbuild["_outdir"]:
        kbuild_out = apkbuild["_outdir"]
    else:
        function_body = pmb.parse.function_body(aport / "APKBUILD", "package")
        kbuild_out = find_kbuild_output_dir(function_body)

    if not kbuild_out:
        kbuild_out = ".output"

    chroot = Chroot.native()

    # Install package dependencies
    depends = pmb.build.get_depends(context, apkbuild)
    pmb.build.init(chroot)
    if arch.cpu_emulation_required():
        depends.append(f"binutils-{arch}")
    pmb.chroot.apk.install(depends, chroot)

    output = pmb.build.output_path(
        arch, apkbuild["pkgname"], apkbuild["pkgver"], apkbuild["pkgrel"]
    )
    message = f"({chroot}) build {output}"
    logging.info(message)

    pmaports_path = pkgrepo_relative_path(aport)[0]

    try:
        run_abuild(context, pkgname, arch, pmaports_path, apkbuild_path, kbuild_out)
    except Exception:
        pmb.helpers.mount.umount_all(Chroot.native() / "mnt/linux")
        raise
    pmb.build.other.index_repo(arch)



# Copyright 2026 Oliver Smith, Paul Adam
# SPDX-License-Identifier: GPL-3.0-or-later
import os

import pmb.chroot
import pmb.chroot.apk
import pmb.chroot.other
import pmb.helpers.apk
import pmb.helpers.mount
import pmb.install.blockdevice
from pmb.core import Chroot, ChrootType
from pmb.helpers import logging
from pmb.helpers.exceptions import CommandFailedError, NonBugError
from pmb.types import Env, RunOutputTypeDefault


def chroot(
    add: str,
    chroot: Chroot,
    chroot_usb: bool,
    command: str,
    install_blockdev: bool,
    output: RunOutputTypeDefault,
    sector_size: int | None,
    user: str,
    xauth: bool,
) -> None:
    if (
        user
        and chroot != Chroot.native()
        and chroot.type not in [ChrootType.BUILDROOT, ChrootType.IMAGE]
    ):
        raise RuntimeError("--user is only supported for native or buildroot_* chroots.")
    if xauth and chroot != Chroot.native():
        raise RuntimeError("--xauth is only supported for native chroot.")

    if chroot.type == ChrootType.IMAGE:
        pmb.chroot.mount(chroot)

    # apk: check minimum version, install packages
    pmb.chroot.apk.check_min_version(chroot)
    if add:
        pmb.chroot.apk.install(add.split(","), chroot)

    pmb.chroot.init(chroot)

    # Xauthority
    env: Env = {}
    if xauth:
        pmb.chroot.other.copy_xauthority(chroot)
        x11_display = os.environ.get("DISPLAY")
        if x11_display is None:
            raise AssertionError("$DISPLAY was unset despite that it should be set at this point")
        env["DISPLAY"] = x11_display
        env["XAUTHORITY"] = "/home/pmos/.Xauthority"

    # Install blockdevice
    if install_blockdev:
        logging.warning(
            "--install-blockdev is deprecated for the chroot command"
            " and will be removed in a future release. If you need this"
            " for some reason, please open an issue on"
            " https://gitlab.postmarketos.org/postmarketOS/pmbootstrap.git"
        )
        size_boot = 128  # 128 MiB
        size_root = 4096  # 4 GiB
        pmb.install.blockdevice.create_and_mount_image(sector_size, size_boot, size_root)

    # Bind mount in sysfs dirs to accessing USB devices (e.g. for running fastboot)
    if chroot_usb:
        for folder in pmb.config.flash_mount_bind:
            pmb.helpers.mount.bind(folder, Chroot.native() / folder)

    pmb.helpers.apk.update_repository_list(chroot.path, user_repository=True)

    try:
        # Run the command as user/root
        if user:
            logging.info(f"({chroot}) % su pmos -c '" + " ".join(command) + "'")
            pmb.chroot.user(command, chroot, output=output, env=env)
        else:
            logging.info(f"({chroot}) % " + " ".join(command))
            pmb.chroot.root(command, chroot, output=output, env=env)
    except CommandFailedError as exception:
        raise NonBugError(exception) from exception



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import pathlib

import pmb.chroot
import pmb.chroot.apk
import pmb.config
from pmb.core import Chroot
from pmb.core.arch import Arch
from pmb.core.context import Context, get_context
from pmb.helpers import logging
from pmb.types import CrossCompile


def init_abuild_minimal(chroot: Chroot = Chroot.native(), build_pkgs: list[str] = []) -> None:
    """Initialize a minimal chroot with abuild where one can do 'abuild checksum'."""
    marker = chroot / "tmp/pmb_chroot_abuild_init_done"
    if os.path.exists(marker):
        return

    if not build_pkgs:
        build_pkgs = pmb.config.build_packages

    pmb.chroot.apk.install(build_pkgs, chroot, build=False)

    # Fix permissions
    pmb.chroot.root(["chown", "root:abuild", "/var/cache/distfiles"], chroot)
    pmb.chroot.root(["chmod", "g+w", "/var/cache/distfiles"], chroot)

    # Add user to group abuild
    pmb.chroot.root(["adduser", "pmos", "abuild"], chroot)

    pathlib.Path(marker).touch()


def init(chroot: Chroot = Chroot.native()) -> bool:
    """Initialize a chroot for building packages with abuild."""
    marker = chroot / "tmp/pmb_chroot_build_init_done"
    if marker.exists():
        return False

    logging.info(f"Initializing {chroot.arch} buildroot")

    # Initialize chroot, install packages
    pmb.chroot.init(Chroot.native())
    pmb.chroot.init(chroot)
    init_abuild_minimal(chroot)

    # Generate package signing keys
    if not os.path.exists(get_context().config.work / "config_abuild/abuild.conf"):
        logging.info(f"({chroot}) generate abuild keys")
        pmb.chroot.user(
            ["abuild-keygen", "-n", "-q", "-a"], chroot, env={"PACKAGER": "pmos <pmos@local>"}
        )

        # Copy package signing key to /etc/apk/keys
        for key in (chroot / "mnt/pmbootstrap/abuild-config").glob("*.pub"):
            key = key.relative_to(chroot.path)
            pmb.chroot.root(["cp", key, "/etc/apk/keys/"], chroot)

    apk_arch = chroot.arch

    # Add apk wrapper that runs native apk and lies about arch
    if apk_arch.cpu_emulation_required() and not (chroot / "usr/local/bin/abuild-apk").exists():
        with (chroot / "tmp/apk_wrapper.sh").open("w") as handle:
            content = f"""
                #!/bin/sh

                # With !pmb:crossdirect, cross compilation is entriely done
                # in QEMU, no /native dir gets mounted inside the foreign arch
                # chroot.
                if ! [ -d /native ]; then
                    exec /usr/bin/abuild-apk "$@"
                fi

                export LD_PRELOAD_PATH=/native/usr/lib:/native/lib
                args=""
                for arg in "$@"; do
                    if [ "$arg" == "--print-arch" ]; then
                        echo "{apk_arch}"
                        exit 0
                    fi
                    args="$args $arg"
                done
                /native/usr/bin/abuild-apk $args
            """
            lines = content.split("\n")[1:]
            for i in range(len(lines)):
                lines[i] = lines[i][16:]
            handle.write("\n".join(lines))
        pmb.chroot.root(["cp", "/tmp/apk_wrapper.sh", "/usr/local/bin/abuild-apk"], chroot)
        pmb.chroot.root(["chmod", "+x", "/usr/local/bin/abuild-apk"], chroot)

    # abuild.conf: Don't clean the build folder after building, so we can
    # inspect it afterwards for debugging
    pmb.chroot.root(["sed", "-i", "-e", "s/^CLEANUP=.*/CLEANUP=''/", "/etc/abuild.conf"], chroot)

    # abuild.conf: Don't clean up installed packages in strict mode, so
    # abuild exits directly when pressing ^C in pmbootstrap.
    pmb.chroot.root(
        ["sed", "-i", "-e", "s/^ERROR_CLEANUP=.*/ERROR_CLEANUP=''/", "/etc/abuild.conf"], chroot
    )

    pathlib.Path(marker).touch()
    return True


def init_compiler(context: Context, depends: list[str], cross: CrossCompile, arch: Arch) -> None:
    arch_str = str(arch)
    cross_pkgs = ["ccache-cross-symlinks", "abuild"]
    if "gcc4" in depends:
        cross_pkgs += ["gcc4-" + arch_str]
    elif "gcc6" in depends:
        cross_pkgs += ["gcc6-" + arch_str]
    elif "clang" in depends or "clang-dev" in depends:
        cross_pkgs += ["clang"]
    else:
        cross_pkgs += ["gcc-" + arch_str, "g++-" + arch_str]
    if cross == CrossCompile.CROSSDIRECT:
        cross_pkgs += ["crossdirect"]
        if "rust" in depends or "cargo" in depends:
            if context.ccache:
                cross_pkgs += ["sccache"]
            # crossdirect for rust installs all build dependencies in the
            # native chroot too, as some of them can be required for building
            # native macros / build scripts
            cross_pkgs += depends

    pmb.chroot.init(Chroot.native())
    pmb.chroot.apk.install(cross_pkgs, Chroot.native())



# Copyright 2024 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from pathlib import Path

from pmb.helpers.exceptions import NonBugError
from pmb.meta import Cache

try:
    # Python >= 3.11
    from tomllib import TOMLDecodeError, load  # novermin
except ImportError:
    # Python < 3.11
    from tomli import TOMLDecodeError, load  # type:ignore[import-not-found,no-redef]


@Cache("path")
def load_toml_file(path: Path) -> dict:
    """Read a toml file into a dict and show the path on error."""
    with open(path, mode="rb") as f:
        try:
            return load(f)
        except TOMLDecodeError as e:
            raise NonBugError(f"{path}: {e}") from e



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from pmb.flasher.init import (
    init as init,
    install_depends as install_depends,
)
from pmb.flasher.run import (
    check_partition_blacklist as check_partition_blacklist,
    run as run,
)
from pmb.flasher.variables import variables as variables



# Copyright 2024 Caleb Connolly
# SPDX-License-Identifier: GPL-3.0-or-later

"""Global runtime context"""

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Literal, overload

from .config import Config


class TimeoutReason(Enum):
    CI_DETECTED = "running in CI based on autodetection (environment variable 'CI' set)"
    TIMEOUT_ARG = "timeout was explicitly set via the --timeout or -t argument"

    def __str__(self) -> str:
        return self.value


@dataclass(frozen=True)
class CommandTimeout:
    length: float
    reason: TimeoutReason


class Context:
    details_to_stdout: bool = False
    quiet: bool = False
    command_timeout: CommandTimeout | None = None
    sudo_timer: bool = False
    force: bool = False
    log: Path

    # assume yes to prompts
    assume_yes: bool = False

    # Operate offline
    offline: bool = False

    # The pmbootstrap subcommand
    command: str = ""

    ## FIXME: build options, should not be here ##
    # disable cross compilation and use QEMU
    cross: bool = False
    no_depends: bool = False
    ignore_depends: bool = False
    ccache: bool = True
    go_mod_cache: bool = False

    config: Config

    def __init__(self, config: Config) -> None:
        self.log = config.work / "log.txt"
        self.config = config


__context: Context


@overload
def get_context(allow_failure: Literal[False] = ...) -> Context: ...


@overload
def get_context(allow_failure: Literal[True] = ...) -> Context | None: ...


def get_context(allow_failure: bool = False) -> Context | None:
    """Get immutable global runtime context."""
    # We must defer this to first call to avoid
    # circular imports.
    if "__context" not in globals():
        if allow_failure:
            return None
        raise RuntimeError("Context not loaded yet")
    return __context


def set_context(context: Context) -> None:
    """Set global runtime context."""
    global __context

    if "__context" in globals():
        raise RuntimeError("Context already loaded")

    __context = context



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import argparse
import contextlib
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import Any, cast

from pmb.core import Config
from pmb.core.arch import Arch
from pmb.helpers.exceptions import NonBugError
from pmb.types import PmbArgs, RunOutputTypeDefault

with contextlib.suppress(ImportError):
    import argcomplete

import pmb.config
import pmb.helpers.args
import pmb.helpers.pmaports

"""This file is about parsing command line arguments passed to pmbootstrap, as
   well as generating the help pages (pmbootstrap -h). All this is done with
   Python's argparse. The parsed arguments get extended and finally stored in
   the "args" variable, which is prominently passed to most functions all
   over the pmbootstrap code base.

   See pmb/helpers/args.py for more information about the args variable.
"""


def toggle_other_boolean_flags(
    *other_destinations: str, value: bool = True
) -> type[argparse.Action]:
    """
    Group several argparse flags to one.
    Sets multiple other_destination to value.

    :param other_destinations: 'the other argument names' str
    :param value: 'the value to set the other_destinations to' bool
    :returns: custom Action
    """

    class SetOtherDestinationsAction(argparse.Action):
        def __init__(self, option_strings: list[str], dest: str, **kwargs: Any) -> None:
            super().__init__(option_strings, dest, nargs=0, const=value, default=value, **kwargs)

        def __call__(
            self,
            parser: argparse.ArgumentParser,
            namespace: argparse.Namespace,
            values: str | Sequence[Any] | None,
            option_string: str | None = None,
        ) -> None:
            for destination in other_destinations:
                setattr(namespace, destination, value)

    return SetOtherDestinationsAction


def arguments_install(subparser: argparse._SubParsersAction) -> None:
    ret = subparser.add_parser(
        "install", help="set up device specific chroot and install to SD card or image file"
    )

    # Other arguments (that don't fit categories below)
    ret.add_argument(
        "--no-sshd", action="store_true", help="do not enable the SSH daemon by default"
    )
    ret.add_argument(
        "--no-firewall", action="store_true", help="do not enable the firewall by default"
    )
    ret.add_argument(
        "--password",
        help="dummy password for automating the"
        " installation - will be handled in PLAIN TEXT during"
        " install and may be logged to the logfile, do not use an"
        " important password!",
    )
    ret.add_argument(
        "--no-cgpt",
        help="do not use cgpt partition table",
        dest="install_cgpt",
        action="store_false",
        default=True,
    )
    ret.add_argument("--zap", help="zap chroots before installing", action="store_true")
    ret.add_argument(
        "--sector-size",
        help="set the sector size for the image file",
        type=int,
        default=None,
        choices=[512, 2048, 4096],
    )

    ret.add_argument(
        "--single-partition",
        action="store_true",
        help="Create a single partition that contains both boot and root files."
        " This can be used on devices that boot without mounting the boot file"
        " system (e.g. Android boot images or fastboot).",
    )

    # Image type
    group_desc = ret.add_argument_group(
        "optional image type",
        "Format of the resulting image. Default is generating a combined image"
        " of the postmarketOS boot and root partitions (--no-split). (If the"
        " device's deviceinfo_flash_method requires separate boot and root"
        " partitions, then --split is the default.) Related:"
        " https://postmarketos.org/partitions",
    )
    group = group_desc.add_mutually_exclusive_group()
    group.add_argument(
        "--no-split",
        help="create combined boot and root image file",
        dest="split",
        action="store_false",
        default=None,
    )
    group.add_argument(
        "--split", help="create separate boot and root image files", action="store_true"
    )
    group.add_argument(
        "--disk",
        "--sdcard",
        help="do not create an image file, instead"
        " write to the given block device (SD card, USB"
        " stick, etc.), for example: '/dev/mmcblk0'",
        metavar="BLOCKDEV",
        type=lambda x: Path(x),
    )
    group.add_argument(
        "--android-recovery-zip",
        help="generate TWRP flashable zip (recommended read: https://postmarketos.org/recoveryzip)",
        action="store_true",
        dest="android_recovery_zip",
    )
    group.add_argument(
        "--no-image", help="do not generate an image", action="store_true", dest="no_image"
    )

    # Image type "--disk" related
    group = ret.add_argument_group("optional image type 'disk' arguments")
    group.add_argument("--rsync", help="update the disk using rsync", action="store_true")

    # Image type "--android-recovery-zip" related
    group = ret.add_argument_group("optional image type 'android-recovery-zip' arguments")
    group.add_argument(
        "--recovery-install-partition",
        default="system",
        help="partition to flash from recovery (e.g. 'external_sd')",
        dest="recovery_install_partition",
    )
    group.add_argument(
        "--recovery-no-kernel",
        help="do not overwrite the existing kernel",
        action="store_false",
        dest="recovery_flash_kernel",
    )

    # Full disk encryption (disabled by default, --no-fde has no effect)
    group = ret.add_argument_group("optional full disk encryption arguments")
    group.add_argument(
        "--fde", help="use full disk encryption", action="store_true", dest="full_disk_encryption"
    )
    group.add_argument("--no-fde", help=argparse.SUPPRESS, action="store_true", dest="no_fde")
    group.add_argument(
        "--cipher",
        help="cryptsetup cipher used to encrypt the the rootfs (e.g. 'aes-xts-plain64')",
        default=pmb.config.defaults["cipher"],
    )
    group.add_argument(
        "--iter-time",
        help="cryptsetup iteration time (in"
        " milliseconds) to use when encrypting the system"
        " partition",
        default=pmb.config.defaults["iter_time"],
    )

    # Packages
    group = ret.add_argument_group(
        "optional packages arguments",
        "Select or deselect packages to be included in the installation.",
    )
    group.add_argument(
        "--add",
        help="comma separated list of packages to be added to the rootfs (e.g. 'vim,gcc')",
        metavar="PACKAGES",
    )
    group.add_argument(
        "--no-base",
        help="do not install postmarketos-base (advanced)",
        action="store_false",
        dest="install_base",
    )
    group.add_argument(
        "--no-recommends",
        dest="install_recommends",
        help="do not install packages listed in _pmb_recommends of the UI pmaports",
        action="store_false",
    )

    # Sparse image
    group_desc = ret.add_argument_group(
        "optional sparse image arguments", "Override deviceinfo_flash_sparse for testing purpose."
    )
    group = group_desc.add_mutually_exclusive_group()
    group.add_argument(
        "--sparse", help="generate sparse image file", default=None, action="store_true"
    )
    group.add_argument(
        "--no-sparse", help="do not generate sparse image file", dest="sparse", action="store_false"
    )

    # Other
    group = ret.add_argument_group("other optional arguments")
    group.add_argument(
        "--no-local-pkgs",
        dest="install_local_pkgs",
        help="do not install locally compiled packages and package signing keys",
        action="store_false",
    )
    group.add_argument(
        "--filesystem", help="root filesystem type", choices=["ext4", "f2fs", "btrfs", "xfs"]
    )


def arguments_export(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser(
        "export",
        help="create convenience symlinks"
        " to generated image files (system, kernel,"
        " initramfs, boot.img, ...)",
    )

    ret.add_argument(
        "export_folder",
        help="export folder, defaults to /tmp/postmarketOS-export",
        default=Path("/tmp/postmarketOS-export"),
        nargs="?",
        type=lambda x: Path(x),
    )
    ret.add_argument(
        "--odin",
        help="odin flashable tar (boot.img/kernel+initramfs only)",
        action="store_true",
        dest="odin_flashable_tar",
    )
    ret.add_argument(
        "--no-install",
        dest="autoinstall",
        default=True,
        help="skip updating kernel/initfs",
        action="store_false",
    )
    return ret


def arguments_sideload(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser(
        "sideload", help="Push packages to a running phone connected over usb or wifi"
    )
    add_packages_arg(ret, nargs="+")
    ret.add_argument(
        "--host",
        help="ip of the device over wifi (defaults to 172.16.42.1)",
        default="172.16.42.1",
    )
    ret.add_argument(
        "--port", help="SSH port of the device over wifi (defaults to 22)", type=int, default=22
    )
    ret.add_argument("--user", help="use a different username than the one set in init")
    ret.add_argument(
        "--arch",
        choices=Arch.supported(),
        help="skip automatic architecture deduction and use the given value",
        type=lambda x: Arch.from_str(x),
    )
    ret.add_argument(
        "--install-key",
        help="install the apk key from this machine if needed",
        action="store_true",
        dest="install_key",
    )
    return ret


def arguments_flasher(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser("flasher", help="flash something to the target device")
    ret.add_argument("--method", help="override flash method", dest="flash_method", default=None)
    sub = ret.add_subparsers(dest="action_flasher")
    sub.required = True

    # Boot, flash kernel
    boot = sub.add_parser("boot", help="boot a kernel once")
    boot.add_argument("--cmdline", help="override kernel commandline")
    flash_kernel = sub.add_parser("flash_kernel", help="flash a kernel")
    for action in [boot, flash_kernel]:
        action.add_argument(
            "--no-install",
            dest="autoinstall",
            default=True,
            help="skip updating kernel/initfs",
            action="store_false",
        )
    flash_kernel.add_argument(
        "--partition",
        default=None,
        help="partition to flash the kernel to (defaults to deviceinfo_flash_*_partition_kernel)",
    )

    # Flash lk2nd
    flash_lk2nd = sub.add_parser(
        "flash_lk2nd",
        help="flash lk2nd, a secondary bootloader needed for various Android devices",
    )
    flash_lk2nd.add_argument(
        "--partition",
        default=None,
        help="partition to flash lk2nd to (defaults to default boot image partition ",
    )

    # Flash rootfs
    flash_rootfs = sub.add_parser(
        "flash_rootfs",
        help="flash the rootfs to a partition on the"
        " device (partition layout does not get"
        " changed)",
    )
    flash_rootfs.add_argument(
        "--partition",
        default=None,
        help="partition to flash the rootfs to (defaults"
        " to deviceinfo_flash_*_partition_rootfs,"
        " 'userdata' on Android may have more"
        " space)",
    )

    # Flash vbmeta
    flash_vbmeta = sub.add_parser(
        "flash_vbmeta",
        help="generate and flash AVB 2.0 image with"
        " disable verification flag set to a"
        " partition on the device (typically called"
        " vbmeta)",
    )
    flash_vbmeta.add_argument(
        "--partition",
        default=None,
        help="partition to flash the vbmeta to (defaults to deviceinfo_flash_*_partition_vbmeta",
    )

    # Flash dtbo
    flash_dtbo = sub.add_parser("flash_dtbo", help="flash dtbo image")
    flash_dtbo.add_argument(
        "--partition",
        default=None,
        help="partition to flash the dtbo to (defaults to deviceinfo_flash_*_partition_dtbo)",
    )

    # Actions without extra arguments
    sub.add_parser("sideload", help="sideload recovery zip")
    sub.add_parser(
        "list_flavors",
        help="list installed kernel flavors" + " inside the device rootfs chroot on this computer",
    )
    sub.add_parser("list_devices", help="show connected devices")

    group = ret.add_argument_group(
        "heimdall options",
        "With heimdall as"
        " flash method, the device automatically"
        " reboots after each flash command. Use"
        " --no-reboot and --resume for multiple"
        " flash actions without reboot.",
    )
    group.add_argument(
        "--no-reboot",
        dest="no_reboot",
        help="don't automatically reboot after flashing",
        action="store_true",
    )
    group.add_argument(
        "--resume",
        dest="resume",
        help="resume flashing after using --no-reboot",
        action="store_true",
    )

    return ret


def arguments_initfs(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser("initfs", help="do something with the initramfs")
    sub = ret.add_subparsers(dest="action_initfs")

    # hook ls
    sub.add_parser("hook_ls", help="list available and installed hook packages")

    # hook add/del
    hook_add = sub.add_parser("hook_add", help="add a hook package")
    hook_del = sub.add_parser("hook_del", help="uninstall a hook package")
    for action in [hook_add, hook_del]:
        action.add_argument(
            "hook",
            help="name of the hook aport, without"
            f" the '{pmb.config.initfs_hook_prefix}' prefix,"
            " for example: 'debug-shell'",
        )

    # ls, build, extract
    sub.add_parser("ls", help="list initramfs contents")
    sub.add_parser("build", help="(re)build the initramfs")
    sub.add_parser("extract", help="extract the initramfs to a temporary folder")

    return ret


def arguments_qemu(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser("qemu")
    ret.add_argument("--cmdline", help="override kernel commandline")
    ret.add_argument("--image-size", help="set rootfs size (e.g. 2048M or 2G)")
    ret.add_argument(
        "--second-storage",
        metavar="IMAGE_SIZE",
        help="add a second storage with the given size (default:"
        " 8G), gets created if it does not exist. Use to"
        " test install from SD to eMMC",
        nargs="?",
        default=None,
        const="8G",
    )
    ret.add_argument("-m", "--memory", type=int, default=1024, help="guest RAM (default: 1024)")
    ret.add_argument("-p", "--port", type=int, default=2222, help="SSH port (default: 2222)")

    ret.add_argument(
        "--no-kvm",
        dest="qemu_kvm",
        default=True,
        action="store_false",
        help="Avoid using hardware-assisted virtualization with KVM even when available (SLOW!)",
    )
    ret.add_argument(
        "--cpu",
        dest="qemu_cpu",
        help="Override emulated QEMU CPU. By default, the host"
        " CPU will be emulated when using KVM and the QEMU"
        " default otherwise (usually a CPU with minimal"
        " features). A useful value is 'max' (emulate all"
        " features that are available), use --cpu help to get a"
        " list of possible values from QEMU.",
    )

    ret.add_argument(
        "--tablet",
        dest="qemu_tablet",
        action="store_true",
        default=False,
        help="Use 'tablet' instead of 'mouse'"
        " input for QEMU. The tablet input device automatically"
        " grabs/releases the mouse when moving in/out of the QEMU"
        " window. (NOTE: For some reason the mouse position is"
        " not reported correctly with this in some cases...)",
    )

    ret.add_argument(
        "--display",
        dest="qemu_display",
        choices=["sdl", "gtk", "none"],
        help="QEMU's display parameter (default: gtk,gl=on)",
        default="gtk",
        nargs="?",
    )
    ret.add_argument(
        "--no-gl",
        dest="qemu_gl",
        default=True,
        action="store_false",
        help="Avoid using GL for accelerating graphics in QEMU  (use software rasterizer, slow!)",
    )
    ret.add_argument(
        "--video",
        dest="qemu_video",
        default="1024x768@60",
        help="Video resolution for QEMU (WidthxHeight@RefreshRate). Default is 1024x768@60.",
    )

    ret.add_argument(
        "--audio",
        dest="qemu_audio",
        choices=["alsa", "pa", "sdl"],
        help="QEMU's audio backend (default: none)",
        default=None,
        nargs="?",
    )

    ret.add_argument(
        "--host-qemu", dest="host_qemu", action="store_true", help="Use the host system's qemu"
    )

    ret.add_argument(
        "--efi",
        action="store_true",
        help="Deprecated: EFI boot is now the default (direct kernel boot has been removed)",
    )

    return ret


def arguments_pkgrel_bump(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser(
        "pkgrel_bump",
        help="increase the pkgrel to"
        " indicate that a package must be rebuilt"
        " because of a dependency change",
    )
    ret.add_argument(
        "--dry",
        action="store_true",
        help="instead of modifying APKBUILDs, exit with >0 when a package would have been bumped",
    )

    # Mutually exclusive: "--auto" or package names
    mode = ret.add_mutually_exclusive_group(required=True)
    mode.add_argument(
        "--auto",
        action="store_true",
        help="all packages which"
        " depend on a library which had an incompatible update"
        " (libraries with a soname bump)",
    )
    mode.add_argument("packages", nargs="*", default=[])
    return ret


def arguments_pkgver_bump(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser(
        "pkgver_bump",
        help="increase the pkgver and reset pkgrel to 0. useful when dealing with metapackages.",
    )

    add_packages_arg(ret, nargs="*", default=[])
    return ret


def arguments_newapkbuild(subparser: argparse._SubParsersAction) -> None:
    """
    Wrapper for Alpine's "newapkbuild" command.

    Most parameters will get directly passed through, and they are defined in
    "pmb/config/__init__.py". That way they can be used here and when passing
    them through in "pmb/helpers/frontend.py". The order of the parameters is
    kept the same as in "newapkbuild -h".
    """
    sub = subparser.add_parser("newapkbuild", help="get a template to package new software")
    sub.add_argument(
        "--folder", help="set postmarketOS aports folder (default: main)", default="main"
    )

    # Passthrough: Strings (e.g. -d "my description")
    for entry in pmb.config.newapkbuild_arguments_strings:
        sub.add_argument(entry[0], dest=entry[1], help=entry[2])

    # Passthrough: Package type switches (e.g. -C for CMake)
    group = sub.add_mutually_exclusive_group()
    for entry in pmb.config.newapkbuild_arguments_switches_pkgtypes:
        group.add_argument(entry[0], dest=entry[1], help=entry[2], action="store_true")

    # Passthrough: Other switches (e.g. -c for copying sample files)
    for entry in pmb.config.newapkbuild_arguments_switches_other:
        sub.add_argument(entry[0], dest=entry[1], help=entry[2], action="store_true")

    # Force switch
    sub.add_argument(
        "-f", dest="force", action="store_true", help="force even if directory already exists"
    )

    # Passthrough: PKGNAME[-PKGVER] | SRCURL
    sub.add_argument(
        "pkgname_pkgver_srcurl",
        metavar="PKGNAME[-PKGVER] | SRCURL",
        help="set either the package name (optionally with the"
        " PKGVER at the end, e.g. 'hello-world-1.0') or the"
        " download link to the source archive",
    )


def arguments_kconfig(subparser: argparse._SubParsersAction) -> None:
    # Allowed architectures
    arch_choices = Arch.supported()

    # Kconfig subparser
    ret = subparser.add_parser("kconfig", help="change or edit kernel configs")
    sub = ret.add_subparsers(dest="action_kconfig")
    sub.required = True

    # "pmbootstrap kconfig check"
    check = sub.add_parser("check", help="check kernel aport config")
    check.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="check all kernels, even the ones that would be ignored by default",
    )
    check.add_argument("--arch", choices=arch_choices, dest="arch", type=lambda x: Arch.from_str(x))
    check.add_argument("--file", help="check a file directly instead of a config in a package")
    check.add_argument(
        "--no-details",
        action="store_false",
        dest="kconfig_check_details",
        help="print one generic error per component instead of"
        " listing each option that needs to be adjusted",
    )
    check.add_argument(
        "-k",
        "--keep-going",
        action="store_true",
        help="continue on errors instead of aborting on the first error",
    )
    check.add_argument(
        "--categories", help="a comma separated list of additional kconfig categories to check"
    )
    add_kernel_arg(check, nargs="*")

    # "pmbootstrap kconfig edit"
    edit = sub.add_parser("edit", help="edit kernel aport config")
    edit.add_argument("--arch", choices=arch_choices, dest="arch", type=lambda x: Arch.from_str(x))
    edit.add_argument("--fragment", help="fragment filename to save changes")
    edit_ui_chooser = edit.add_mutually_exclusive_group()
    edit_ui_chooser.add_argument(
        "-x",
        dest="xconfig",
        action="store_true",
        help="use xconfig rather than menuconfig for kernel configuration",
    )
    edit_ui_chooser.add_argument(
        "-n",
        dest="nconfig",
        action="store_true",
        help="use nconfig rather than menuconfig for kernel configuration",
    )
    add_kernel_arg(edit, nargs=1)

    # "pmbootstrap kconfig migrate"
    migrate = sub.add_parser(
        "migrate",
        help="Migrate kconfig from older version to "
        "newer. Internally runs 'make oldconfig', "
        "which asks question for every new kernel "
        "config option.",
    )
    migrate.add_argument(
        "--arch", choices=arch_choices, dest="arch", type=lambda x: Arch.from_str(x)
    )
    add_kernel_arg(migrate, nargs=1)

    generate = sub.add_parser("generate", help="generate kernel config from fragments")
    generate.add_argument(
        "--arch", choices=arch_choices, dest="arch", type=lambda x: Arch.from_str(x)
    )
    add_kernel_arg(generate, nargs=1)


def arguments_repo_missing(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser(
        "repo_missing",
        help="list all packages + depends from pmaports for building the repository (used by bpo)",
    )
    ret.add_argument(
        "--arch", choices=Arch.supported(), default=Arch.native(), type=lambda x: Arch.from_str(x)
    )
    # Deprecated argument that is currently kept so pmbootstrap can be called
    # the same way for repo_missing by bpo with pmbv2 and pmbv3. Once we drop
    # support for pmbv2 in bpo (can do that after v24.06 is EOL), we can adjust
    # bpo to not use --built and remove this parameter from pmbootstrap.
    ret.add_argument(
        "--built",
        action="store_true",
        help=argparse.SUPPRESS,
    )
    return ret


def arguments_test(subparser: argparse._SubParsersAction) -> None:
    test = subparser.add_parser("test", help="Internal pmbootstrap test tools")
    sub = test.add_subparsers(dest="action_test", required=True)
    sub.add_parser("apkindex_parse_all", help="parse all APKINDEX files")


def arguments_status(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    return subparser.add_parser("status", help="show a config and pmaports overview")


def arguments_netboot(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser("netboot", help="launch nbd server with pmOS rootfs")
    sub = ret.add_subparsers(dest="action_netboot")
    sub.required = True

    start = sub.add_parser("serve", help="start nbd server")
    start.add_argument("--replace", action="store_true", help="replace stored netboot image")

    return ret


def arguments_ci(subparser: argparse._SubParsersAction) -> argparse.ArgumentParser:
    ret = subparser.add_parser(
        "ci",
        help="run continuous integration scripts locally of git repo in current directory",
    )
    script_args = ret.add_mutually_exclusive_group()
    script_args.add_argument("-a", "--all", action="store_true", help="run all scripts")
    script_args.add_argument("-f", "--fast", action="store_true", help="run fast scripts only")
    ret.add_argument(
        "scripts",
        nargs="*",
        metavar="script",
        help="name of the CI script to run, depending on the git repository",
    )
    return ret


def package_completer(
    prefix: str,
    action: str,
    parser: argparse.ArgumentParser | None = None,
    parsed_args: list[str] | None = None,
) -> set[str]:
    return {package for package in pmb.helpers.pmaports.get_list() if package.startswith(prefix)}


def kernel_completer(
    prefix: str,
    action: str,
    parser: argparse.ArgumentParser | None = None,
    parsed_args: list[str] | None = None,
) -> list[str]:
    """:returns: matched linux-* packages, with linux-* prefix and without"""
    ret: list[str] = []

    # Full package name, starting with "linux-"
    if (len("linux-") < len(prefix) and prefix.startswith("linux-")) or "linux-".startswith(prefix):
        ret += package_completer(prefix, action, parser, parsed_args)

    # Kernel name without "linux-"
    packages = package_completer(f"linux-{prefix}", action, parser, parsed_args)
    ret += [package.replace("linux-", "", 1) for package in packages]

    return ret


def add_packages_arg(
    subparser: argparse.ArgumentParser, name: str = "packages", *args: str, **kwargs: Any
) -> None:
    arg = subparser.add_argument(name, *args, **kwargs)
    if "argcomplete" in sys.modules:
        arg.completer = package_completer  # type: ignore[attr-defined]


def add_kernel_arg(
    subparser: argparse.ArgumentParser, name: str = "package", nargs: int | str = "?"
) -> None:
    arg = subparser.add_argument(
        name, nargs=nargs, help="kernel package (e.g. linux-postmarketos-allwinner)"
    )
    if "argcomplete" in sys.modules:
        arg.completer = kernel_completer  # type: ignore[attr-defined]


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="pmbootstrap")
    arch_native = Arch.native()
    arch_choices = Arch.supported()

    # Other
    parser.add_argument("-V", "--version", action="version", version=pmb.__version__)
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        type=lambda x: Path(x),
        default=pmb.config.defaults["config"],
        help="path to pmbootstrap_v3.cfg file (default in ~/.config/)",
    )
    parser.add_argument(
        "-mp",
        "--mirror-pmOS",
        dest="deprecated_mp",
        help=argparse.SUPPRESS,
        action="append",
    )
    parser.add_argument(
        "-m",
        "--mirror-alpine",
        dest="deprecated_m",
        help=argparse.SUPPRESS,
    )
    parser.add_argument("-j", "--jobs", help="parallel jobs when compiling")
    parser.add_argument(
        "-E",
        "--extra-space",
        help="specify an integer with the amount of additional"
        "space to allocate to the image in MB (default"
        " 0)",
    )
    parser.add_argument(
        "-B",
        "--boot-size",
        help="specify an integer with your preferred boot"
        "partition size on target machine in MB (default"
        f" {Config.get_default('boot_size')})",
    )
    parser.add_argument(
        "-p",
        "--aports",
        help="postmarketos aports (pmaports) path",
        type=lambda x: [Path(p.strip()) for p in x.split(",")],
    )
    parser.add_argument(
        "-t",
        "--timeout",
        help="seconds after which processes get killed that stopped writing any output (default: "
        "no timeout unless the environment variable 'CI' is set, in which case it is 900)",
        type=float,
    )
    parser.add_argument(
        "-w",
        "--work",
        type=lambda x: Path(x).absolute(),
        help="folder where all data gets stored (chroots, caches, built packages)",
    )
    parser.add_argument(
        "-y",
        "--assume-yes",
        help="Assume 'yes' to all"
        " question prompts. WARNING: this option will"
        " cause normal 'are you sure?' prompts to be"
        " disabled!",
        action="store_true",
    )
    parser.add_argument(
        "--as-root",
        help="Allow running as root (not"
        " recommended, may screw up your work folders"
        " directory permissions!)",
        dest="as_root",
        action="store_true",
    )
    parser.add_argument(
        "-o",
        "--offline",
        help="Do not attempt to update the package index files",
        action="store_true",
    )

    # Compiler
    parser.add_argument(
        "--no-ccache", action="store_false", dest="ccache", help="do not cache the compiled output"
    )
    parser.add_argument(
        "--no-cross",
        action="store_false",
        dest="cross",
        help="disable cross compiler, build only with QEMU and gcc (slow!)",
    )

    # Logging
    parser.add_argument("-l", "--log", dest="log", default=None, help="path to log file")
    parser.add_argument(
        "--details-to-stdout",
        dest="details_to_stdout",
        help="print details (e.g. build output) to stdout, instead of writing to the log",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="write even more to the logfiles (this may reduce performance)",
    )
    parser.add_argument(
        "-q", "--quiet", dest="quiet", action="store_true", help="do not output any log messages"
    )

    # Actions
    sub = parser.add_subparsers(title="action", dest="action")
    sub.add_parser("shutdown", help="umount, unregister binfmt")
    sub.add_parser(
        "index",
        help="re-index all repositories with custom built"
        " packages (do this after manually removing package files)",
    )
    sub.add_parser(
        "work_migrate",
        help="run this before using pmbootstrap"
        " non-interactively to migrate the"
        " work folder version on demand",
    )
    arguments_repo_missing(sub)
    arguments_kconfig(sub)
    arguments_export(sub)
    arguments_sideload(sub)
    arguments_netboot(sub)
    arguments_flasher(sub)
    arguments_initfs(sub)
    arguments_qemu(sub)
    arguments_pkgrel_bump(sub)
    arguments_pkgver_bump(sub)
    arguments_newapkbuild(sub)
    arguments_test(sub)
    arguments_status(sub)
    arguments_ci(sub)

    # Action: init
    init = sub.add_parser("init", help="initialize config file")
    init.add_argument(
        "--shallow-initial-clone",
        help="do a shallow clone if pmaports has to be cloned. Primarily useful to speed up scripts.",
        action="store_true",
    )

    # Action: log
    log = sub.add_parser("log", help="follow the pmbootstrap logfile")
    log.add_argument("-n", "--lines", type=int, default=60, help="count of initial output lines")
    log.add_argument("-c", "--clear", help="clear the log", action="store_true", dest="clear_log")

    # Action: zap
    zap = sub.add_parser("zap", help="safely delete chroot folders")
    zap.add_argument(
        "--dry",
        action="store_true",
        help="instead of actually deleting anything, print out what would have been deleted",
    )
    zap.add_argument("-hc", "--http", action="store_true", help="also delete http cache")
    zap.add_argument(
        "-d", "--distfiles", action="store_true", help="also delete downloaded source tarballs"
    )
    zap.add_argument(
        "-p",
        "--pkgs-local",
        action="store_true",
        dest="pkgs_local",
        help="also delete *all* locally compiled packages",
    )
    zap.add_argument(
        "-m",
        "--pkgs-local-mismatch",
        action="store_true",
        dest="pkgs_local_mismatch",
        help="also delete locally compiled packages without existing aport of same version",
    )
    zap.add_argument(
        "-n", "--netboot", action="store_true", help="also delete stored images for netboot"
    )
    zap.add_argument(
        "-o",
        "--pkgs-online-mismatch",
        action="store_true",
        dest="pkgs_online_mismatch",
        help="also delete outdated packages from online mirrors"
        " (that have been downloaded to the apk cache)",
    )
    zap.add_argument("-r", "--rust", action="store_true", help="also delete rust related caches")

    zap_all_delete_args = [
        "http",
        "distfiles",
        "pkgs_local",
        "pkgs_local_mismatch",
        "netboot",
        "pkgs_online_mismatch",
        "rust",
    ]
    zap_all_delete_args_print = [arg.replace("_", "-") for arg in zap_all_delete_args]
    zap.add_argument(
        "-a",
        "--all",
        action=toggle_other_boolean_flags(*zap_all_delete_args),
        help=f"delete everything, equivalent to: --{' --'.join(zap_all_delete_args_print)}",
    )

    # Action: stats
    stats = sub.add_parser("stats", help="show ccache stats")
    stats.add_argument(
        "--arch", default=arch_native, choices=arch_choices, type=lambda x: Arch.from_str(x)
    )

    # Action: update
    update = sub.add_parser("update", help="update all existing APKINDEX files")
    update.add_argument(
        "--arch",
        default=None,
        choices=arch_choices,
        help="only update a specific architecture",
        type=lambda x: Arch.from_str(x),
    )
    update.add_argument(
        "--non-existing",
        action="store_true",
        help="do not only update the existing APKINDEX files, but all of them",
        dest="non_existing",
    )

    # Action: build_init / chroot
    build_init = sub.add_parser(
        "build_init",
        help="initialize build environment (usually you do not need to call this)",
    )
    chroot = sub.add_parser("chroot", help="start shell in chroot")
    chroot.add_argument(
        "--add",
        help="build/install comma separated list of packages in the chroot before entering it",
    )
    chroot.add_argument("--user", help="run the command as user, not as root", action="store_true")
    chroot.add_argument(
        "--output",
        choices=list(RunOutputTypeDefault),
        type=RunOutputTypeDefault.from_string,
        help="how the output of the"
        " program should be handled, choose from: 'log',"
        " 'stdout', 'interactive', 'tui' (default),"
        " 'background'. Details: pmb/helpers/run_core.py",
        default=RunOutputTypeDefault.TUI,
    )
    chroot.add_argument(
        "--image",
        help="Mount the rootfs image and treat it like a normal chroot.",
        action="store_true",
    )
    chroot.add_argument(
        "--usb",
        help="Make USB devices accessible inside the chroot.",
        action="store_true",
        dest="chroot_usb",
    )
    chroot.add_argument(
        "command",
        default=["sh", "-i"],
        help="command to execute inside the chroot. default: sh",
        nargs="*",
    )
    chroot.add_argument(
        "-x",
        "--xauth",
        action="store_true",
        help="Copy .Xauthority and set environment variables,"
        " so X11 applications can be started (native"
        " chroot only)",
    )
    chroot.add_argument(
        "-i",
        "--install-blockdev",
        action="store_true",
        help="Create a sparse image file and mount it as"
        " /dev/install, just like during the"
        " installation process.",
    )
    for action in [build_init, chroot]:
        suffix = action.add_mutually_exclusive_group()
        if action == chroot:
            suffix.add_argument(
                "-r", "--rootfs", action="store_true", help="Chroot for the device root file system"
            )
        suffix.add_argument(
            "-b",
            "--buildroot",
            nargs="?",
            const="device",
            choices={"device"} | {str(a) for a in arch_choices},
            help="Chroot for building packages, defaults to device architecture",
        )
        suffix.add_argument(
            "-s",
            "--suffix",
            default=None,
            help="Specify any chroot suffix, defaults to 'native'",
        )

    # Action: install
    arguments_install(sub)

    # Action: checksum
    checksum = sub.add_parser("checksum", help="update aport checksums")
    checksum.add_argument(
        "--verify",
        action="store_true",
        help="download"
        " sources and verify that the checksums of the"
        " APKBUILD match, instead of updating them",
    )
    checksum_changed = checksum.add_mutually_exclusive_group(required=False)
    checksum_changed.add_argument(
        "--changed",
        action="store_true",
        help="update checksums of all packages that have unstaged or uncommitted changes",
        required=False,
    )
    checksum_changed.add_argument("packages", nargs="*", default=[])

    # Action: aportgen
    aportgen = sub.add_parser(
        "aportgen",
        help="generate a postmarketOS specific package build recipe (aport/APKBUILD)",
    )
    aportgen_fork_alpine = aportgen.add_mutually_exclusive_group()
    aportgen_fork_alpine.add_argument(
        "-a",
        "--fork-alpine",
        help="fork the alpine upstream package",
        action="store_true",
        dest="fork_alpine",
    )
    aportgen_fork_alpine.add_argument(
        "-r",
        "--fork-alpine-retain-branch",
        help="fork the alpine upstream, but don't change branch to match the current channel",
        action="store_true",
        dest="fork_alpine_retain_branch",
    )
    add_packages_arg(aportgen, nargs="+")

    # Action: build
    build = sub.add_parser("build", help="create a package for a specific architecture")
    build.add_argument(
        "--arch",
        choices=arch_choices,
        default=None,
        help="CPU architecture to build for (default: "
        f"{arch_native} or first available architecture in"
        " APKBUILD)",
        type=lambda x: Arch.from_str(x),
    )
    build.add_argument("--force", action="store_true", help="even build if not necessary")
    build.add_argument(
        "--strict",
        action="store_true",
        help="(slower) zap and"
        " install only required depends when building, to"
        " detect dependency errors",
    )
    build.add_argument(
        "--src",
        help="override source used to build the"
        " package with a local folder (the APKBUILD must"
        " expect the source to be in $builddir, so you might"
        " need to adjust it)",
        nargs=1,
    )
    build.add_argument(
        "-i",
        "--ignore-depends",
        action="store_true",
        help="only build and install makedepends from an"
        " APKBUILD, ignore the depends (old behavior). This is"
        " faster for device packages for example, because then"
        " you don't need to build and install the kernel. But"
        " it is incompatible with how Alpine's abuild handles"
        " it.",
        dest="ignore_depends",
    )
    build.add_argument(
        "-n",
        "--no-depends",
        action="store_true",
        help="never build dependencies, abort instead",
        dest="no_depends",
    )
    build.add_argument(
        "--go-mod-cache",
        action="store_true",
        default=None,
        help="for go packages: Usually they should bundle the"
        " dependency sources instead of downloading them"
        " at build time. But if they don't (e.g. with"
        " pmbootstrap build --src), then this option can"
        " be used to let GOMODCACHE point into"
        " pmbootstrap's work dir to only download"
        " dependencies once. (default: true with --src,"
        " false otherwise)",
    )
    build.add_argument(
        "--no-go-mod-cache",
        action="store_false",
        dest="go_mod_cache",
        default=None,
        help="don't set GOMODCACHE",
    )
    build.add_argument(
        "--envkernel",
        action="store_true",
        help="Create an apk package from the build output of"
        " a kernel compiled locally on the host or with envkernel.sh.",
    )
    add_packages_arg(build, nargs="+")

    # Action: deviceinfo_parse
    deviceinfo_parse = sub.add_parser("deviceinfo_parse")
    deviceinfo_parse.add_argument("devices", nargs="*")
    deviceinfo_parse.add_argument(
        "--kernel",
        help="the kernel to select (for"
        " device packages with multiple kernels),"
        " e.g. 'downstream', 'mainline'",
        dest="deviceinfo_parse_kernel",
        metavar="KERNEL",
    )

    # Action: apkbuild_parse
    apkbuild_parse = sub.add_parser("apkbuild_parse")
    add_packages_arg(apkbuild_parse, nargs="*")

    # Action: apkindex_parse
    apkindex_parse = sub.add_parser("apkindex_parse")
    apkindex_parse.add_argument("apkindex_path", type=lambda x: Path(x))
    add_packages_arg(apkindex_parse, "package", nargs="?")

    # Action: config
    config = sub.add_parser("config", help="get and set pmbootstrap options")
    config.add_argument(
        "-r",
        "--reset",
        action="store_true",
        help="Reset config options with the given name to it's default.",
    )
    config.add_argument(
        "name",
        nargs="?",
        help="variable name, one of: " + ", ".join(sorted(Config.keys())),
        choices=Config.keys(),
        metavar="name",
    )
    config.add_argument("value", nargs="?", help="set variable to value")

    # Action: bootimg_analyze
    bootimg_analyze = sub.add_parser(
        "bootimg_analyze", help="Extract all the information from an existing boot.img"
    )
    bootimg_analyze.add_argument("path", help="path to the boot.img", type=lambda x: Path(x))
    bootimg_analyze.add_argument(
        "--force", "-f", action="store_true", help="force even if the file seems to be invalid"
    )

    # Action: pull
    sub.add_parser(
        "pull", help="update all git repositories that pmbootstrap cloned (pmaports, etc.)"
    )

    if "argcomplete" in sys.modules:
        argcomplete.autocomplete(parser, always_complete_options="long")

    return parser


def arguments() -> PmbArgs:
    # FIXME: It would be nice to not use cast here, but I don't know what else we could do.
    args = cast(PmbArgs, get_parser().parse_args())

    if getattr(args, "fork_alpine_retain_branch", False):
        # fork_alpine_retain_branch largely matches the behaviour of fork_alpine, so
        # just set fork_alpine here to reduce repetition.
        args.fork_alpine = args.fork_alpine_retain_branch

    pmb.helpers.args.init(args)

    if getattr(args, "deprecated_mp", None) or getattr(args, "deprecated_m", None):
        raise NonBugError(
            "Arguments --mirror-pmOS and --mirror-alpine have been removed. See docs/mirrors.md"
            " regarding how to set mirrors now."
        )

    if getattr(args, "go_mod_cache", None) is None:
        gomodcache = bool(getattr(args, "src", None))
        args.go_mod_cache = gomodcache

    return args



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import socket
from contextlib import closing

import pmb.chroot
import pmb.helpers.mount
import pmb.install.losetup
from pmb.core import Chroot, ChrootType
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging


def kill_adb() -> None:
    """Kill adb daemon if it's running."""
    port = 5038
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(2)
        if sock.connect_ex(("127.0.0.1", port)) == 0:
            pmb.chroot.root(["adb", "-P", str(port), "kill-server"])


def kill_sccache() -> None:
    """
    Kill sccache daemon if it's running. Unlike ccache it automatically spawns
    a daemon when you call it and exits after some time of inactivity.
    """
    port = 4226
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(2)
        if sock.connect_ex(("127.0.0.1", port)) == 0:
            pmb.chroot.root(["sccache", "--stop-server"])


def shutdown_cryptsetup_device(name: str) -> None:
    """:param name: cryptsetup device name, usually "pm_crypt" in pmbootstrap"""
    if not (Chroot.native() / "dev/mapper" / name).exists():
        return
    pmb.chroot.apk.install(["cryptsetup"], Chroot.native())
    status = pmb.chroot.root(["cryptsetup", "status", name], output_return=True, check=False)
    if not status:
        logging.warning(
            "WARNING: Failed to run cryptsetup to get the status"
            " for " + name + ", assuming it is not mounted"
            " (shutdown fails later if it is)!"
        )
        return

    if status.startswith("/dev/mapper/" + name + " is active."):
        pmb.chroot.root(["cryptsetup", "luksClose", name])
    elif status.startswith("/dev/mapper/" + name + " is inactive."):
        # When "cryptsetup status" fails, the device is not mounted and we
        # have a left over file (#83)
        pmb.chroot.root(["rm", "/dev/mapper/" + name])
    else:
        raise RuntimeError("Failed to parse 'cryptsetup status' output!")


def shutdown(only_install_related: bool = False) -> None:
    # Stop daemons
    kill_adb()
    kill_sccache()

    chroot = Chroot.native()

    # Umount installation-related paths (order is important!)
    pmb.helpers.mount.umount_all(chroot / "mnt/install")
    shutdown_cryptsetup_device("pm_crypt")

    # Umount all losetup mounted images
    pmb.install.losetup.detach_all()

    # Remove "in-pmbootstrap" marker from all chroots. This marker indicates
    # that pmbootstrap has set up all mount points etc. to run programs inside
    # the chroots, but we want it gone afterwards (e.g. when the chroot
    # contents get copied to a rootfs / installer image, or if creating an
    # android recovery zip from its contents).
    for marker in get_context().config.work.glob("chroot_*/in-pmbootstrap"):
        pmb.helpers.run.root(["rm", marker])

    # Umount device rootfs and installer chroots
    if only_install_related:
        for chroot_type in [ChrootType.ROOTFS, ChrootType.INSTALLER]:
            chroot = Chroot(chroot_type, get_context().config.device)
            if chroot.path.exists():
                pmb.helpers.mount.umount_all(chroot.path)
        return

    # Umount all folders inside work dir
    # The folders are explicitly iterated over, so folders symlinked inside
    # work dir get umounted as well (used in test_pkgrel_bump.py, #1595)
    for path in get_context().config.work.glob("*"):
        pmb.helpers.mount.umount_all(path)

    # Clean up the rest
    for arch in Arch.supported():
        if arch.cpu_emulation_required():
            pmb.chroot.binfmt.unregister(arch)

    logging.debug("Shutdown complete")



# Copyright 2026 Hugo Posnic
# SPDX-License-Identifier: GPL-3.0-or-later
import sys
from typing import NoReturn

import pmb.helpers.status


def status() -> NoReturn:
    pmb.helpers.status.print_status()

    # Do not print the DONE! line
    sys.exit(0)



# Copyright 2024 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import logging
from pathlib import Path

import pmb.config
from pmb.core.arch import Arch
from pmb.core.pkgrepo import pkgrepo_default_path
from pmb.helpers.toml import NonBugError, load_toml_file
from pmb.meta import Cache


@Cache()
def get_path() -> Path:
    """
    Get the kconfigcheck.toml from current pmaports branch if it exists, or
    as fallback the v24.06 version shipped with pmbootstrap.
    """
    ret = pkgrepo_default_path() / "kconfigcheck.toml"
    if ret.exists():
        return ret

    logging.info(
        "NOTE: couldn't find kconfigcheck.toml in pmaports dir, using"
        " the version from postmarketOS v24.06"
    )
    return pmb.config.pmb_src / "pmb/data/kconfigcheck.toml"


def sanity_check(toml: dict) -> None:
    """Ensure the kconfigcheck.toml file has the expected structure."""
    path = get_path()

    if "aliases" not in toml:
        raise RuntimeError(f"{path}: missing [aliases] section")

    for alias in toml["aliases"]:
        for category in toml["aliases"][alias]:
            if not category.startswith("category:"):
                raise RuntimeError(
                    f"{path}: alias {alias}: all categories must start with 'category:'!"
                )

    for section in toml:
        if section == "aliases":
            continue
        if not all(cat.startswith("category:") for cat in section.split()):
            raise RuntimeError(f"{path}: unexpected section: {section}")
        for versions in toml[section]:
            for arches in toml[section][versions]:
                if arches == "all":
                    continue
                if not isinstance(toml[section][versions][arches], dict):
                    raise RuntimeError(f"{path}: {section} is missing architecture information")
                for arch in arches.split(" "):
                    _ = Arch.from_str(arch)


@Cache("categories")
def read_categories(categories: list[str]) -> dict[str, dict]:
    """Read multiple categories (including aliases) from kconfigcheck.toml."""
    toml = load_toml_file(get_path())
    sanity_check(toml)

    # Potentially resolve category alias
    # category: exists
    real_categories: dict[str, bool] = {}
    for category in categories:
        if category in toml["aliases"]:
            resolved_aliases = [c.split(":", 1)[1] for c in toml["aliases"][category]]
            real_categories |= dict.fromkeys(resolved_aliases, False)
            logging.debug(f"kconfigcheck: read_categories: '{category}' -> {resolved_aliases}")
        else:
            real_categories[category] = False

    ret = {}

    for key in toml:
        # Keys may contain multiple space-separated category:name entries, which all need
        # to be satisfied for a section to be considered.
        if key.startswith("category:"):
            # This must be a kconfigcheck section
            required_categories = [c.split(":", 1)[1] for c in key.split()]
            all_requirements_met = True

            for required_category in required_categories:
                if required_category in real_categories:
                    # This category exists!
                    real_categories[required_category] = True
                else:
                    # We still keep going through categories since we still want
                    # to be able to error on nonexisting categories
                    all_requirements_met = False

            if all_requirements_met:
                logging.debug(f"kconfigcheck: section {key} has all requirements met")
                ret[key] = toml[key]

    # Make sure that all specified categories actually exist in the TOML
    for category, exists in real_categories.items():
        if not exists:
            raise RuntimeError(f"{get_path()}: couldn't find {category}")

    return ret


def get_generic_kconfig() -> dict[str, dict]:
    """Reads the contents of kconfig-generic.toml and returns the parsed TOML."""
    path = Path(pkgrepo_default_path(), "kconfig-generic.toml")
    try:
        return load_toml_file(path)
    except NonBugError as exception:
        raise NonBugError(
            f"kconfig-generic.toml not found, please update your pmaports checkout with 'pmbootstrap pull'. ({exception})"
        ) from exception



# Copyright 2023 Oliver Smith
# Copyright 2024 Stefan Hansson
# SPDX-License-Identifier: GPL-3.0-or-later

import pmb.parse.deviceinfo
from pmb.core.context import get_context
from pmb.flasher.frontend import flash_lk2nd, kernel, list_flavors, rootfs, sideload
from pmb.helpers import logging


def flasher(
    action: str,
    autoinstall: bool,
    cmdline: str | None,
    flash_method: str,
    no_reboot: bool | None,
    partition: str | None,
    resume: bool | None,
) -> None:
    context = get_context()
    device = context.config.device
    deviceinfo = pmb.parse.deviceinfo()
    method = flash_method or deviceinfo.flash_method

    if method == "none" and action in ["boot", "flash_kernel", "flash_rootfs", "flash_lk2nd"]:
        logging.info("This device doesn't support any flash method.")
        return

    if action in ["boot", "flash_kernel"]:
        kernel(
            deviceinfo,
            method,
            action == "boot",
            autoinstall,
            cmdline=cmdline,
            no_reboot=no_reboot,
            partition=partition,
            resume=resume,
        )
    elif action == "flash_rootfs":
        rootfs(
            deviceinfo,
            method,
            cmdline=cmdline,
            no_reboot=no_reboot,
            partition=partition,
            resume=resume,
        )
    elif action == "flash_vbmeta":
        logging.info("(native) flash vbmeta.img with verity disabled flag")
        pmb.flasher.run(
            deviceinfo,
            method,
            "flash_vbmeta",
            cmdline=cmdline,
            no_reboot=no_reboot,
            partition=partition,
            resume=resume,
        )
    elif action == "flash_dtbo":
        logging.info("(native) flash dtbo image")
        pmb.flasher.run(
            deviceinfo,
            method,
            "flash_dtbo",
            cmdline=cmdline,
            no_reboot=no_reboot,
            partition=partition,
            resume=resume,
        )
    elif action == "flash_lk2nd":
        flash_lk2nd(
            deviceinfo,
            method,
            cmdline=cmdline,
            no_reboot=no_reboot,
            partition=partition,
            resume=resume,
        )
    elif action == "list_flavors":
        list_flavors(device)
    elif action == "list_devices":
        pmb.flasher.run(
            deviceinfo,
            method,
            "list_devices",
            cmdline=cmdline,
            no_reboot=no_reboot,
            partition=partition,
            resume=resume,
        )
    elif action == "sideload":
        sideload(
            deviceinfo,
            method,
            cmdline=cmdline,
            no_reboot=no_reboot,
            partition=partition,
            resume=resume,
        )



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from collections.abc import Generator
from pathlib import Path

import pmb.chroot.apk
import pmb.config
import pmb.parse._apkbuild
from pmb.core import Chroot
from pmb.core.pkgrepo import pkgrepo_iglob
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError


def path_to_hook_name(path: Path) -> str:
    return path.name[len(pmb.config.initfs_hook_prefix) :]


def list_chroot(suffix: Chroot, remove_prefix: bool = True) -> list[str]:
    ret = []
    prefix = pmb.config.initfs_hook_prefix
    for pkgname in pmb.chroot.apk.installed(suffix):
        if pkgname.startswith(prefix):
            if remove_prefix:
                ret.append(pkgname[len(prefix) :])
            else:
                ret.append(pkgname)
    return ret


def list_hook_paths() -> Generator[Path, None, None]:
    return pkgrepo_iglob(f"*/{pmb.config.initfs_hook_prefix}*")


def list_hook_packages() -> list[str]:
    return [path_to_hook_name(path) for path in list_hook_paths()]


def ls(suffix: Chroot) -> None:
    hooks_chroot = list_chroot(suffix)

    for hook_path in list_hook_paths():
        hook_desc = pmb.parse._apkbuild.apkbuild(hook_path)["pkgdesc"]
        hook = path_to_hook_name(hook_path)
        line = f"* {hook}: {hook_desc} ({'' if hook in hooks_chroot else 'not '}installed)"
        logging.info(line)


def add(hook: str, suffix: Chroot) -> None:
    if hook not in list_hook_packages():
        raise NonBugError(
            "Invalid hook name! Run 'pmbootstrap initfs hook_ls' to get a list of all hooks."
        )
    prefix = pmb.config.initfs_hook_prefix
    pmb.chroot.apk.install([f"{prefix}{hook}"], suffix)


def delete(hook: str, suffix: Chroot) -> None:
    if hook not in list_chroot(suffix):
        raise NonBugError("There is no such hook installed!")
    prefix = pmb.config.initfs_hook_prefix
    pmb.helpers.apk.run(["del", f"{prefix}{hook}"], suffix)


def update(suffix: Chroot) -> None:
    """Rebuild and update all hooks that are out of date"""
    pmb.chroot.apk.install(list_chroot(suffix, False), suffix)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import collections
from enum import IntEnum

"""
In order to stay as compatible to Alpine's apk as possible, this code
is heavily based on:

https://gitlab.alpinelinux.org/alpine/apk-tools/-/blob/5d796b567819ce91740fcdea7cbafecbda65d8f3/src/version.c
"""


class Token(IntEnum):
    """C equivalent: enum PARTS"""

    INVALID = -1
    DIGIT_OR_ZERO = 0
    DIGIT = 1
    LETTER = 2
    SUFFIX = 3
    SUFFIX_NO = 4
    REVISION_NO = 5
    END = 6


def next_token(previous: Token, rest: str) -> tuple[Token, str]:
    """
    Parse the next token in the rest of the version string, we're
    currently looking at.

    We do *not* get the value of the token, or advance the rest string
    beyond the whole token that is what the get_token() function does
    (see below).

    :param previous: the token before
    :param rest: of the version string
    :returns: (next, rest) next is the upcoming token, rest is the
              input "rest" string with one leading '.', '_' or '-'
              character removed (if there was any).

    C equivalent: next_token()
    """
    next = Token.INVALID
    char = rest[:1]

    # Tokes, which do not change rest
    if not len(rest):
        next = Token.END
    elif previous in [Token.DIGIT, Token.DIGIT_OR_ZERO] and char.islower():
        next = Token.LETTER
    elif previous == Token.LETTER and char.isdigit():
        next = Token.DIGIT
    elif previous == Token.SUFFIX and char.isdigit():
        next = Token.SUFFIX_NO

    # Tokens, which remove the first character of rest
    else:
        if char == ".":
            next = Token.DIGIT_OR_ZERO
        elif char == "_":
            next = Token.SUFFIX
        elif rest.startswith("-r"):
            next = Token.REVISION_NO
            rest = rest[1:]
        elif char == "-":
            next = Token.INVALID
        rest = rest[1:]

    # Validate current token
    # Check if the transition from previous to current is valid
    if next < previous and not (
        (next == Token.DIGIT_OR_ZERO and previous == Token.DIGIT)
        or (next == Token.SUFFIX and previous == Token.SUFFIX_NO)
        or (next == Token.DIGIT and previous == Token.LETTER)
    ):
        next = Token.INVALID
    return (next, rest)


def parse_suffix(rest: str) -> tuple[str, int, bool]:
    """
    Cut off the suffix of rest (which is now at the beginning of the
    rest variable, but regarding the whole version string, it is a
    suffix), and return a value integer (so it can be compared later,
    "beta" > "alpha" etc).

    :param rest: what is left of the version string that we are
                 currently parsing, starts with a "suffix" value
                 (see below for valid suffixes).
    :returns: (rest, value, invalid_suffix)
              - rest: is the input "rest" string without the suffix
              - value: is a signed integer (negative for pre-,
              positive for post-suffixes).
              - invalid_suffix: is true, when rest does not start
              with anything from the suffixes variable.

    C equivalent: get_token(), case TOKEN_SUFFIX
    """
    name_suffixes = collections.OrderedDict(
        [
            ("pre", ["alpha", "beta", "pre", "rc"]),
            ("post", ["cvs", "svn", "git", "hg", "p"]),
        ]
    )

    for name, suffixes in name_suffixes.items():
        for i, suffix in enumerate(suffixes):
            if not rest.startswith(suffix):
                continue
            rest = rest[len(suffix) :]
            value = i
            if name == "pre":
                value = value - len(suffixes)
            return (rest, value, False)
    return (rest, 0, True)


def get_token(previous: Token, rest: str) -> tuple[Token, int, str]:
    """
    This function does three things:
    * get the next token
    * get the token value
    * cut-off the whole token from rest

    :param previous: the token before
    :param rest: of the version string
    :returns: (next, value, rest) next is the new token string,
              value is an integer for comparing, rest is the rest of the
              input string.

    C equivalent: get_token()
    """
    # Set defaults
    value = 0
    next = Token.INVALID
    invalid_suffix = False

    # Bail out if at the end
    if not len(rest):
        return (Token.END, 0, rest)

    # Cut off leading zero digits
    if previous == Token.DIGIT_OR_ZERO and rest.startswith("0"):
        while rest.startswith("0"):
            rest = rest[1:]
            value -= 1
        next = Token.DIGIT

    # Add up numeric values
    elif previous in [Token.DIGIT_OR_ZERO, Token.DIGIT, Token.SUFFIX_NO, Token.REVISION_NO]:
        for i in range(len(rest)):
            while len(rest) and rest[0].isdigit():
                value *= 10
                value += int(rest[i])
                rest = rest[1:]

    # Append chars or parse suffix
    elif previous == Token.LETTER:
        value = ord(rest[0])
        rest = rest[1:]
    elif previous == Token.SUFFIX:
        (rest, value, invalid_suffix) = parse_suffix(rest)

    # Invalid previous token
    else:
        value = -1

    # Get the next token (for non-leading zeros)
    if not len(rest):
        next = Token.END
    elif next == Token.INVALID and not invalid_suffix:
        (next, rest) = next_token(previous, rest)

    return (next, value, rest)


def validate(version: str) -> bool:
    """
    Check whether one version string is valid.

    :param version: full version string
    :returns: True when the version string is valid

    C equivalent: apk_version_validate()
    """
    current = Token.DIGIT
    rest = version
    while current != Token.END:
        (current, _, rest) = get_token(current, rest)
        if current == Token.INVALID:
            return False
    return True


def compare(a_version: str, b_version: str, fuzzy: bool = False) -> int:
    """
    Compare two versions A and B to find out which one is higher, or if
    both are equal.

    :param a_version: full version string A
    :param b_version: full version string B
    :param fuzzy: treat version strings, which end in different token
                  types as equal

    :returns:
        (a <  b): -1
        (a == b):  0
        (a >  b):  1

    C equivalent: apk_version_compare_blob_fuzzy()
    """
    # Defaults
    a_token = Token.DIGIT
    b_token = Token.DIGIT
    a_value = 0
    b_value = 0
    a_rest = a_version
    b_rest = b_version

    # Parse A and B one token at a time, until one string ends, or the
    # current token has a different type/value
    while a_token == b_token and a_token not in [Token.END, Token.INVALID] and a_value == b_value:
        (a_token, a_value, a_rest) = get_token(a_token, a_rest)
        (b_token, b_value, b_rest) = get_token(b_token, b_rest)

    # Compare the values inside the last tokens
    if a_value < b_value:
        return -1
    if a_value > b_value:
        return 1

    # Equal: When tokens are the same strings, or when the value
    # is the same and fuzzy compare is enabled
    if a_token == b_token or fuzzy:
        return 0

    # Leading version components and their values are equal, now the
    # non-terminating version is greater unless it's a suffix
    # indicating pre-release
    if a_token == Token.SUFFIX:
        (a_token, a_value, a_rest) = get_token(a_token, a_rest)
        if a_value < 0:
            return -1
    if b_token == Token.SUFFIX:
        (b_token, b_value, b_rest) = get_token(b_token, b_rest)
        if b_value < 0:
            return 1

    # Compare the token value (e.g. digit < letter)
    if a_token > b_token:
        return -1
    if a_token < b_token:
        return 1

    # The tokens are not the same, but previous checks revealed that it
    # is equal anyway (e.g. "1.0" == "1").
    return 0


"""
Convenience functions below are not modeled after apk's version.c.
"""


def check_string(a_version: str, rule: str) -> bool:
    """
    Compare a version against a check string. This is used in "pmbootstrap
    kconfig check", to only require certain options if the pkgver is in a
    specified range (#1795).

    :param a_version: "3.4.1"
    :param rule: ">=1.0.0"
    :returns: True if a_version matches rule, false otherwise.

    """
    # Operators and the expected returns of compare(a,b)
    operator_results = {">=": [1, 0], "<": [-1]}

    # Find the operator
    b_version = None
    expected_results = None
    for operator in operator_results:
        if rule.startswith(operator):
            b_version = rule[len(operator) :]
            expected_results = operator_results[operator]
            break

    # No operator found
    if not b_version:
        raise RuntimeError(
            "Could not find operator in '" + rule + "'. You"
            " probably need to adjust check_string() in"
            " pmb/parse/version.py."
        )

    # Compare
    result = compare(a_version, b_version)
    return not expected_results or result in expected_results



# Copyright 2023 Anjandev Momi
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import shutil
from functools import lru_cache


@lru_cache
def which_sudo() -> str | None:
    """
    Return a command required to run commands as root, if any.

    Find whether sudo, doas, or run0 is installed for commands that require root.
    Allows user to override preferred sudo with PMB_SUDO env variable.
    """
    if os.getuid() == 0:
        return None

    supported_sudos = ["doas", "sudo", "run0"]

    user_set_sudo = os.getenv("PMB_SUDO")
    if user_set_sudo is not None:
        if shutil.which(user_set_sudo) is None:
            raise RuntimeError(
                "PMB_SUDO environmental variable is set to"
                f" {user_set_sudo} but pmbootstrap cannot find"
                " this command on your system."
            )
        return user_set_sudo

    for sudo in supported_sudos:
        if shutil.which(sudo) is not None:
            return sudo

    raise RuntimeError(
        "Can't find sudo, doas, or run0 required to run pmbootstrap."
        " Please install sudo, doas, run0, or specify your own sudo"
        " with the PMB_SUDO environmental variable."
    )



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Save, read, verify workdir state related information in $WORK/workdir.cfg,
for example the init dates of the chroots. This is not saved in
pmbootstrap_v3.cfg, because pmbootstrap_v3.cfg is not tied to a specific work
dir.
"""

import configparser
import os
import time
from typing import overload

import pmb.config
import pmb.config.pmaports
from pmb.core import Chroot, ChrootType
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError


def chroot_save_init(suffix: Chroot) -> None:
    """Save the chroot initialization data in $WORK/workdir.cfg."""
    # Read existing cfg
    cfg = configparser.ConfigParser()
    path = get_context().config.work / "workdir.cfg"
    if os.path.isfile(path):
        cfg.read(path)

    # Create sections
    for key in ["chroot-init-dates", "chroot-channels"]:
        if key not in cfg:
            cfg[key] = {}

    # Update sections
    channel = pmb.config.pmaports.read_config()["channel"]
    cfg["chroot-channels"][str(suffix)] = channel
    cfg["chroot-init-dates"][str(suffix)] = str(int(time.time()))

    # Write back
    with open(path, "w") as handle:
        cfg.write(handle)


@overload
def chroots_outdated() -> list[Chroot]: ...


@overload
def chroots_outdated(chroot: Chroot) -> bool: ...


def chroots_outdated(chroot: Chroot | None = None) -> bool | list[Chroot]:
    """
    Check if init dates from workdir.cfg indicate that any chroot is
    outdated.

    :param suffix: only check a specific chroot suffix

    :returns: A list of all outdated chroots if chroot is None, if a specific
              chroot is given, instead it returns True if the chroot is outdated,
              False otherwise
    """
    # Skip if workdir.cfg doesn't exist
    path = get_context().config.work / "workdir.cfg"
    if not os.path.exists(path):
        return False if chroot else []

    cfg = configparser.ConfigParser()
    cfg.read(path)
    key = "chroot-init-dates"
    if key not in cfg:
        return False if chroot else []

    outdated: list[Chroot] = []
    date_outdated = time.time() - pmb.config.chroot_outdated
    for cfg_suffix in cfg[key]:
        if chroot and cfg_suffix != str(chroot):
            continue
        date_init = int(cfg[key][cfg_suffix])
        if date_init <= date_outdated:
            if chroot:
                return True
            outdated.append(Chroot.from_str(cfg_suffix))
    return False if chroot else outdated


def chroot_check_channel(chroot: Chroot) -> bool:
    """
    Check the chroot channel against the current channel. Returns
    True if the chroot should be zapped (both that it needs zapping and
    the user has auto_zap_misconfigured_chroots enabled), False otherwise.
    """
    if chroot.type == ChrootType.IMAGE:
        return False

    config = get_context().config
    path = config.work / "workdir.cfg"
    msg_again = (
        "Run 'pmbootstrap zap' to delete your chroots and try again."
        " To do this automatically, run 'pmbootstrap config"
        " auto_zap_misconfigured_chroots yes'."
    )
    msg_unknown = f"Could not figure out on which release channel the '{chroot}' chroot is."
    if not os.path.exists(path):
        raise RuntimeError(f"{msg_unknown} {msg_again}")

    cfg = configparser.ConfigParser()
    cfg.read(path)
    key = "chroot-channels"
    if key not in cfg or str(chroot) not in cfg[key]:
        raise RuntimeError(f"{msg_unknown} {msg_again}")

    channel = pmb.config.pmaports.read_config()["channel"]
    channel_cfg = cfg[key][str(chroot)]
    msg = (
        f"Chroot '{chroot}' is for the '{channel_cfg}' channel,"
        f" but you are on the '{channel}' channel."
    )

    if channel != channel_cfg:
        if config.auto_zap_misconfigured_chroots.enabled():
            if config.auto_zap_misconfigured_chroots.noisy():
                logging.info(msg)
                logging.info(
                    "Automatically zapping since auto_zap_misconfigured_chroots is enabled."
                )
                logging.info(
                    "NOTE: You can silence this message with 'pmbootstrap"
                    " config auto_zap_misconfigured_chroots silently'"
                )
            else:
                logging.debug(f"{msg} Zapping chroot.")
            return True
        raise NonBugError(f"{msg} {msg_again}")

    return False


def clean() -> bool | None:
    """
    Remove obsolete data data from workdir.cfg.

    :returns: None if workdir does not exist,
        True if config was rewritten,
        False if config did not change
    """
    # Skip if workdir.cfg doesn't exist
    path = get_context().config.work / "workdir.cfg"
    if not os.path.exists(path):
        return None

    # Read
    cfg = configparser.ConfigParser()
    cfg.read(path)

    # Remove entries for deleted chroots
    changed = False
    for key in ["chroot-init-dates", "chroot-channels"]:
        if key not in cfg:
            continue
        for suffix_str in cfg[key]:
            suffix = Chroot.from_str(suffix_str)
            if suffix.path.exists():
                continue
            changed = True
            del cfg[key][suffix_str]

    # Write back
    if changed:
        with path.open("w") as handle:
            cfg.write(handle)

    return changed



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
# PYTHON_ARGCOMPLETE_OK
import os
import sys
import traceback
from pathlib import Path
from typing import TYPE_CHECKING

from pmb.helpers.exceptions import BuildFailedError, NonBugError, PackagingError

if TYPE_CHECKING:
    from pmb.types import PmbArgs

from . import config, parse
from .commands import run_command
from .config import (
    init as config_init,
    require_programs,
)
from .core import Chroot, Config
from .core.context import get_context
from .helpers import logging, mount, other, status

# pmbootstrap version
__version__ = "3.10.2"

# Python version check
# === CHECKLIST FOR UPGRADING THE REQUIRED PYTHON VERSION ===
# * .ci/vermin.sh
# * README.md
# * docs/usage.rst
# * pmb/__init__.py (you are here)
# * pyproject.toml
# * when upgrading to python 3.11: pmb/helpers/toml.py and remove this line
version = sys.version_info
if version < (3, 10):
    print("You need at least Python 3.10 to run pmbootstrap")
    print("(You are running it with Python " + str(version.major) + "." + str(version.minor) + ")")
    sys.exit()


def print_log_hint() -> None:
    context = get_context(allow_failure=True)
    if context and context.details_to_stdout:
        return
    log = context.log if context else Config().work / "log.txt"
    # Hints about the log file (print to stdout only)
    log_hint = "Run 'pmbootstrap log' for details."
    if not os.path.exists(log):
        log_hint += (
            " Alternatively you can use '--details-to-stdout' to get more"
            " output, e.g. 'pmbootstrap --details-to-stdout init'."
        )
    print()
    print(log_hint)


def main() -> int:
    # Wrap everything to display nice error messages

    args: PmbArgs
    try:
        # Parse arguments, set up logging
        args = parse.arguments()
        context = get_context()
        os.umask(0o22)

        # Store script invocation command
        os.environ["PMBOOTSTRAP_CMD"] = sys.argv[0]

        # Sanity checks
        other.check_grsec()
        if not args.as_root and os.geteuid() == 0:
            raise RuntimeError("Do not run pmbootstrap as root!")

        # Check for required programs (and find their absolute paths)
        require_programs()

        # Initialize or require config
        if args.action == "init":
            config_init.frontend(args)
            return 0
        elif not os.path.exists(args.config):
            if args.config != config.defaults["config"]:
                raise NonBugError(f"Couldn't find file passed with --config: {args.config}")
            old_config = (
                Path(os.environ.get("XDG_CONFIG_HOME") or os.path.expanduser("~/.config"))
                / "pmbootstrap.cfg"
            )
            if os.path.exists(old_config):
                raise NonBugError(
                    f"Thanks for upgrading to pmbootstrap {__version__}!"
                    " The config file format has changed, please generate a new config with"
                    " 'pmbootstrap init'."
                )
            raise NonBugError(
                "Run 'pmbootstrap init' first to generate a config file (or use --config)."
            )
        elif not os.path.exists(context.config.work):
            raise NonBugError("Work path not found, please run 'pmbootstrap init' to create it.")

        # Migrate work folder if necessary
        if args.action not in ["shutdown", "zap", "log"]:
            other.migrate_work_folder()

        # Run the function with the action's name (in pmb/helpers/frontend.py)
        if args.action:
            run_command(args)
        else:
            logging.info("Run pmbootstrap -h for usage information.")

        # Still active notice
        if mount.ismount(Chroot.native() / "dev"):
            logging.info("NOTE: chroot is still active (use 'pmbootstrap shutdown' as necessary)")
        logging.info("DONE!")

    except KeyboardInterrupt:
        print("\nCaught KeyboardInterrupt, exiting …")
        sys.exit(130)  # SIGINT(2) + 128

    except NonBugError as exception:
        logging.error(f"ERROR: {exception}")
        return 2

    except BuildFailedError as exception:
        logging.error(f"ERROR: {exception}")
        print_log_hint()
        return 3

    except PackagingError as exception:
        logging.error(f"ERROR: {exception}")
        print_log_hint()
        print("""
This error likely originates from a packaging issue. Please ensure that your
local pmaports clone is up-to-date with 'pmbootstrap pull'. If the issue still
persists, try zapping all chroots with 'pmbootstrap zap --all' and try again.

Should that still not work, please open an issue in the pmaports (not
pmbootstrap) issue tracker including the pmbootstrap log.""")
        return 4

    except Exception as e:
        # Dump log to stdout when args (and therefore logging) init failed
        can_print_status = get_context(allow_failure=True) is not None
        if "args" not in locals():
            import logging as pylogging

            pylogging.getLogger().setLevel(logging.DEBUG)
            can_print_status = False

        logging.info("ERROR: " + str(e))
        logging.info("See also: <https://postmarketos.org/troubleshooting>")
        logging.debug(traceback.format_exc())

        print_log_hint()
        print()
        print("Before you report this error, ensure that pmbootstrap is up to date.")
        print(
            "Find the latest version here: https://gitlab.postmarketos.org/postmarketOS/pmbootstrap/-/tags"
        )
        print(f"Your version: {__version__}")
        if can_print_status:
            status.print_status()
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())
